#include "g_local.h"
#include "m_player.h"

void ClientUserinfoChanged (edict_t *ent, char *userinfo);

void SP_misc_teleporter_dest (edict_t *ent);

void OpenClassMenu (edict_t *ent, pmenuhnd_t *p);
void BeSoldier (edict_t *ent, pmenuhnd_t *p);
void BeEnforcer (edict_t *ent, pmenuhnd_t *p);
void BeTank (edict_t *ent, pmenuhnd_t *p);
void BeMedic (edict_t *ent, pmenuhnd_t *p);
void BeGunner (edict_t *ent, pmenuhnd_t *p);
void BeGladiator (edict_t *ent, pmenuhnd_t *p);
void BeSupertank (edict_t *ent, pmenuhnd_t *p);
void BeFlyer (edict_t *ent, pmenuhnd_t *p);
void BeBerserk (edict_t *ent, pmenuhnd_t *p);
void BeIcarus (edict_t *ent, pmenuhnd_t *p);
void BeChick (edict_t *ent, pmenuhnd_t *p);
void OpenGroundMenu (edict_t *ent, pmenuhnd_t *p);
void OpenFlyMenu (edict_t *ent, pmenuhnd_t *p);
void GoBack (edict_t *ent, pmenuhnd_t *p);
void EndObserverMode_Walking(edict_t* ent);
void EndObserverMode_Flying(edict_t* ent);
void EndObserverMode_Tread(edict_t* ent);
void GoToGround2Menu (edict_t *ent, pmenuhnd_t *p);
void BeBoss2 (edict_t *ent, pmenuhnd_t *p);
void BeBrain (edict_t *ent, pmenuhnd_t *p);
void BeMutant (edict_t *ent, pmenuhnd_t *p);
void BeJorg (edict_t *ent, pmenuhnd_t *p);
void BeMakron (edict_t *ent, pmenuhnd_t *p);
void BeFloater (edict_t *ent, pmenuhnd_t *p);
void BeShark (edict_t *ent, pmenuhnd_t *p);
void OpenSwimMenu (edict_t *ent, pmenuhnd_t *p);
// Pmenu stuff

// Note that the pmenu entries are duplicated
// this is so that a static set of pmenu entries can be used
// for multiple clients and changed without interference
// note that arg will be freed when the menu is closed, it must be allocated memory
pmenuhnd_t *PMenu_Open(edict_t *ent, pmenu_t *entries, int cur, int num, void *arg)
{
	pmenuhnd_t *hnd;
	pmenu_t *p;
	int i;

	if (!ent->client)
		return NULL;

	if (ent->client->menu) {
		gi.dprintf("warning, ent already has a menu\n");
		PMenu_CloseIt(ent);
	}

	hnd = malloc(sizeof(*hnd));

	hnd->arg = arg;
	hnd->entries = malloc(sizeof(pmenu_t) * num);
	memcpy(hnd->entries, entries, sizeof(pmenu_t) * num);
	// duplicate the strings since they may be from static memory
	for (i = 0; i < num; i++)
		if (entries[i].text)
			hnd->entries[i].text = strdup(entries[i].text);

	hnd->num = num;

	if (cur < 0 || !entries[cur].SelectFunc) {
		for (i = 0, p = entries; i < num; i++, p++)
			if (p->SelectFunc)
				break;
	} else
		i = cur;

	if (i >= num)
		hnd->cur = -1;
	else
		hnd->cur = i;

	ent->client->showscores = true;
	ent->client->inmenu = true;
	ent->client->menu = hnd;

	PMenuDoUpdate(ent);
	gi.unicast (ent, true);

	return hnd;
}

// only use on pmenu's that have been called with PMenu_Open
void PMenu_UpdateEntry(pmenu_t *entry, const char *text, int align, SelectFunc_t SelectFunc)
{
	if (entry->text)
		free(entry->text);
	entry->text = strdup(text);
	entry->align = align;
	entry->SelectFunc = SelectFunc;
}

void PMenu_Update(edict_t *ent)
{
	if (!ent->client->menu) {
		gi.dprintf("warning:  ent has no menu\n");
		return;
	}

	if (level.time - ent->client->menutime >= 0.2) {
		// been a second or more since last update, update now
		PMenuDoUpdate(ent);
		gi.unicast (ent, true);
		ent->client->menutime = level.time;
		ent->client->menudirty = false;
	}
	ent->client->menutime = level.time + 0.2;
	ent->client->menudirty = true;
}

void PMenu_Next(edict_t *ent)
{
	pmenuhnd_t *hnd;
	int i;
	pmenu_t *p;

	if (!ent->client->menu) {
		gi.dprintf("warning:  ent has no menu\n");
		return;
	}

	hnd = ent->client->menu;

	if (hnd->cur < 0)
		return; // no selectable entries

	i = hnd->cur;
	p = hnd->entries + hnd->cur;
	do {
		i++, p++;
		if (i == hnd->num)
			i = 0, p = hnd->entries;
		if (p->SelectFunc)
			break;
	} while (i != hnd->cur);

	hnd->cur = i;

	PMenu_Update(ent);
}

void PMenu_Prev(edict_t *ent)
{
	pmenuhnd_t *hnd;
	int i;
	pmenu_t *p;

	if (!ent->client->menu) {
		gi.dprintf("warning:  ent has no menu\n");
		return;
	}

	hnd = ent->client->menu;

	if (hnd->cur < 0)
		return; // no selectable entries

	i = hnd->cur;
	p = hnd->entries + hnd->cur;
	do {
		if (i == 0) {
			i = hnd->num - 1;
			p = hnd->entries + i;
		} else
			i--, p--;
		if (p->SelectFunc)
			break;
	} while (i != hnd->cur);

	hnd->cur = i;

	PMenu_Update(ent);
}

void PMenu_Select(edict_t *ent)
{
	pmenuhnd_t *hnd;
	pmenu_t *p;

	if (!ent->client->menu) {
		gi.dprintf("warning:  ent has no menu\n");
		return;
	}

	hnd = ent->client->menu;

	if (hnd->cur < 0)
		return; // no selectable entries

	p = hnd->entries + hnd->cur;

	if (p->SelectFunc)
		p->SelectFunc(ent, hnd);
}

void CloseThisMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
}

void GoBackMain (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
}

pmenu_t mainmenu[] = {
	{ "*Main Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Join the game",	PMENU_ALIGN_LEFT, OpenClassMenu }, 
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },   
	{ NULL,				PMENU_ALIGN_LEFT, NULL }, 
	{ "Read the readme",PMENU_ALIGN_LEFT, NULL }, 
	{ "for any help.",	PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Close",			PMENU_ALIGN_LEFT, CloseThisMenu }
};

pmenu_t classtypemenu[] = {
	{ "*ClassType Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Select a Classtype",				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL }, 
	{ "Ground",				PMENU_ALIGN_LEFT, OpenGroundMenu },
	{ "Flying",				PMENU_ALIGN_LEFT, OpenFlyMenu },
	{ "Swimming",				PMENU_ALIGN_LEFT, OpenSwimMenu },   
	{ NULL,				PMENU_ALIGN_LEFT, NULL }, 
	{ "Read the readme",PMENU_ALIGN_LEFT, NULL }, 
	{ "for any help.",	PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Close",			PMENU_ALIGN_LEFT, CloseThisMenu }
};

pmenu_t swimmenu[] = {
	{ "*SwimType Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Barracuda Shark",				PMENU_ALIGN_LEFT, BeShark },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Back",			PMENU_ALIGN_LEFT, OpenGroundMenu }
};

pmenu_t ground2[] = {
	{ "*GroundType Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Iron Maiden",				PMENU_ALIGN_LEFT, BeChick },
	{ "Brain",				PMENU_ALIGN_LEFT, BeBrain},
	{ "Mutant",				PMENU_ALIGN_LEFT, BeMutant },
	{ "Parasite",				PMENU_ALIGN_LEFT, NULL },
	{ "JORG",				PMENU_ALIGN_LEFT, BeJorg },
	{ "Makron",				PMENU_ALIGN_LEFT, BeMakron },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Back",			PMENU_ALIGN_LEFT, OpenGroundMenu }
};

pmenu_t groundmenu[] = {
	{ "*GroundType Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Soldier",				PMENU_ALIGN_LEFT, BeSoldier },
	{ "Enforcer",				PMENU_ALIGN_LEFT, BeEnforcer },
	{ "Tank",				PMENU_ALIGN_LEFT, BeTank },   
	{ "Medic",				PMENU_ALIGN_LEFT, BeMedic },
	{ "Gunner",				PMENU_ALIGN_LEFT, BeGunner },
	{ "Gladiator",PMENU_ALIGN_LEFT, BeGladiator }, 
	{ "Supertank",	PMENU_ALIGN_LEFT, BeSupertank },
	{ "Berserk",				PMENU_ALIGN_LEFT, BeBerserk },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Next",			PMENU_ALIGN_LEFT, GoToGround2Menu }
};

pmenu_t flymenu[] = {
	{ "*FlyType Menu",		PMENU_ALIGN_CENTER, NULL },
	{ "StroggDM",		PMENU_ALIGN_CENTER, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Flyer",				PMENU_ALIGN_LEFT, BeFlyer },
	{ "Icarus",				PMENU_ALIGN_LEFT, BeIcarus },
	{ "Floater",				PMENU_ALIGN_LEFT, BeFloater },
	{ "Boss2",				PMENU_ALIGN_LEFT, BeBoss2 },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ NULL,				PMENU_ALIGN_LEFT, NULL },
	{ "Back",				PMENU_ALIGN_LEFT, GoBack },
	{ "Close",			PMENU_ALIGN_LEFT, CloseThisMenu }
};

void OpenGroundMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, groundmenu, -1, sizeof(groundmenu) / sizeof(pmenu_t), NULL);
}

void GoToGround2Menu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, ground2, -1, sizeof(ground2) / sizeof(pmenu_t), NULL);
}
void GoBack (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, mainmenu, -1, sizeof(mainmenu) / sizeof(pmenu_t), NULL);
}

void OpenFlyMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, flymenu, -1, sizeof(flymenu) / sizeof(pmenu_t), NULL);
}

void OpenSwimMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, swimmenu, -1, sizeof(swimmenu) / sizeof(pmenu_t), NULL);
}

void OpenClassMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, classtypemenu, -1, sizeof(classtypemenu) / sizeof(pmenu_t), NULL);
}
void OpenMainMenus (edict_t *ent)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, mainmenu, -1, sizeof(mainmenu) / sizeof(pmenu_t), NULL);
}
void OpenMainMenu (edict_t *ent, pmenuhnd_t *p)
{
	PMenu_CloseIt (ent);
	PMenu_Open(ent, mainmenu, -1, sizeof(mainmenu) / sizeof(pmenu_t), NULL);
}
void BeSoldier (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 1;
    EndObserverMode_Walking(ent);
}
void BeEnforcer (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 2;
    EndObserverMode_Walking(ent);
}
void BeTank (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 3;
    EndObserverMode_Walking(ent);
}
void BeMedic (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 4;
    EndObserverMode_Walking(ent);
}
void BeGunner (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 5;
    EndObserverMode_Walking(ent);
}
void BeGladiator (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 6;
    EndObserverMode_Walking(ent);
}
void BeSupertank (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 7;
    EndObserverMode_Tread(ent);
}
void BeFlyer (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 8;
    EndObserverMode_Flying(ent);
}
void BeShark (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 17;
    EndObserverMode_Walking(ent);
}
void BeFloater (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 16;
    EndObserverMode_Flying(ent);
}
void BeBerserk (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 9;
    EndObserverMode_Walking(ent);
}
void BeChick (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 10;
    EndObserverMode_Walking(ent);
}
void BeBoss2 (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 11;
    EndObserverMode_Flying(ent);
}
void BeBrain (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 12;
    EndObserverMode_Walking(ent);
}
void BeMutant (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 13;
    EndObserverMode_Walking(ent);
}
void BeJorg (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 18;
    EndObserverMode_Walking(ent);
}
void BeMakron (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 19;
    EndObserverMode_Walking(ent);
}
void BeIcarus (edict_t *ent, pmenuhnd_t *p)
{
    ent->client->resp.class = 15;
    EndObserverMode_Flying(ent);
}
//
// Gross, ugly, disgustuing hack section
//

// this function is an ugly as hell hack to fix some map flaws
//
// the coop spawn spots on some maps are SNAFU.  There are coop spots
// with the wrong targetname as well as spots with no name at all
//
// we use carnal knowledge of the maps to fix the coop spot targetnames to match
// that of the nearest named single player spot

/*we get silly velocity-effects when we are on ground and try to
  accelerate, so lift us a little bit if possible*/
qboolean Jet_AvoidGround( edict_t *ent )
{
  vec3_t         new_origin;
  trace_t trace;
  qboolean       success;
 
  /*Check if there is enough room above us before we change origin[2]*/
  new_origin[0] = ent->s.origin[0];
  new_origin[1] = ent->s.origin[1];
  new_origin[2] = ent->s.origin[2] + 0.5;
  trace = gi.trace( ent->s.origin, ent->mins, ent->maxs, new_origin, ent, MASK_MONSTERSOLID );
 
  if ( success=(trace.plane.normal[2]==0) )    /*no ceiling?*/
    ent->s.origin[2] += 0.5;                   /*then make sure off ground*/
 
  return success;
}
 
 
/*This function returns true if the jet is activated
  (surprise, surprise)*/
qboolean Jet_Active( edict_t *ent )
{
  return ( ent->client->Jet_framenum >= level.framenum );
}

/*If a player dies with activated jetpack this function will be called
  and produces a little explosion*/
void Jet_BecomeExplosion( edict_t *ent, int damage )
{
  int    n;
 
  gi.WriteByte( svc_temp_entity );
  gi.WriteByte( TE_EXPLOSION1 );   /*TE_EXPLOSION2 is possible too*/
  gi.WritePosition( ent->s.origin );
  gi.multicast( ent->s.origin, MULTICAST_PVS );
  gi.sound( ent, CHAN_BODY, gi.soundindex("misc/udeath.wav"), 1, ATTN_NORM, 0 );
 
  /*throw some gib*/
  for ( n=0; n<4; n++ )     ThrowGib( ent, "models/objects/gibs/sm_meat/tris.md2", damage, GIB_ORGANIC );   ThrowClientHead( ent, damage );   ent->takedamage = DAMAGE_NO;
 
}
 
 
/*The lifting effect is done through changing the origin, it
  gives the best results. Of course its a little dangerous because
  if we dont take care, we can move into solid*/
void Jet_ApplyLifting( edict_t *ent )
{
  float          delta;
  vec3_t new_origin;
  trace_t trace;
  int            time = 24;     /*must be >0, time/10 = time in sec for a
                                 complete cycle (up/down)*/
  float          amplitude = 2.0;
 
  /*calculate the z-distance to lift in this step*/
  delta = sin( (float)((level.framenum%time)*(360/time))/180*M_PI ) * amplitude;
  delta = (float)((int)(delta*8))/8; /*round to multiples of 0.125*/
 
  VectorCopy( ent->s.origin, new_origin );
  new_origin[2] += delta;
 
  if( VectorLength(ent->velocity) == 0 )
  {
     /*i dont know the reason yet, but there is some floating so we
       have to compensate that here (only if there is no velocity left)*/
     new_origin[0] -= 0.125;
     new_origin[1] -= 0.125;
     new_origin[2] -= 0.125;
  }
 
  /*before we change origin, its important to check that we dont go
    into solid*/
  trace = gi.trace( ent->s.origin, ent->mins, ent->maxs, new_origin, ent, MASK_MONSTERSOLID );
  if ( trace.plane.normal[2] == 0 )
    VectorCopy( new_origin, ent->s.origin );
}
 
 
/*This function applys some sparks to your jetpack, this part is
  exactly copied from Muce's and SumFuka's JetPack-tutorial and does a
  very nice effect.*/
void Jet_ApplySparks ( edict_t *ent )
{
  vec3_t  forward, right;
  vec3_t  pack_pos, jet_vector;
 
  AngleVectors(ent->client->v_angle, forward, right, NULL);
  VectorScale (forward, -7, pack_pos);
  VectorAdd (pack_pos, ent->s.origin, pack_pos);
  pack_pos[2] += (ent->viewheight);
  VectorScale (forward, -50, jet_vector);
 
  gi.WriteByte (svc_temp_entity);
  gi.WriteByte (TE_SPARKS);
  gi.WritePosition (pack_pos);
  gi.WriteDir (jet_vector);
  gi.multicast (pack_pos, MULTICAST_PVS);
}
 
 
/*if the angle of the velocity vector is different to the viewing
  angle (flying curves or stepping left/right) we get a dotproduct
  which is here used for rolling*/
void Jet_ApplyRolling( edict_t *ent, vec3_t right )
{
  float roll,
        value = 0.05,
        sign = -1;    /*set this to +1 if you want to roll contrariwise*/
 
  roll = DotProduct( ent->velocity, right ) * value * sign;
  ent->client->kick_angles[ROLL] = roll;
}
 
 
/*Now for the main movement code. The steering is a lot like in water, that
  means your viewing direction is your moving direction. You have three
  direction Boosters: the big Main Booster and the smaller up-down and
  left-right Boosters.
  There are only 2 adds to the code of the first tutorial: the Jet_next_think
  and the rolling.
  The other modifications results in the use of the built-in quake functions,
  there is no change in moving behavior (reinventing the wheel is a lot of
  "fun" and a BIG waste of time ;-))*/
void Jet_ApplyJet( edict_t *ent, usercmd_t *ucmd )
{
  float  direction;
  vec3_t acc;
  vec3_t forward, right;
  int    i;
 
  /*clear gravity so we dont have to compensate it with the Boosters*/
  ent->client->ps.pmove.gravity = 0;
 
  /*calculate the direction vectors dependent on viewing direction
    (length of the vectors forward/right is always 1, the coordinates of
    the vectors are values of how much youre looking in a specific direction
    [if youre looking up to the top, the x/y values are nearly 0 the
    z value is nearly 1])*/
  AngleVectors( ent->client->v_angle, forward, right, NULL );
 
  /*Run jet only 10 times a second so movement dont depends on fps
    because ClientThink is called as often as possible
    (fps<10 still is a problem ?)*/   if ( ent->client->Jet_next_think <= level.framenum )
  {
    ent->client->Jet_next_think = level.framenum + 1;
 
    /*clear acceleration-vector*/
    VectorClear( acc );
 
    /*if we are moving forward or backward add MainBooster acceleration
      (60)*/
    if ( ucmd->forwardmove )
    {
      /*are we accelerating backward or forward?*/
      direction = (ucmd->forwardmove<0) ? -1.0 : 1.0;        /*add the acceleration for each direction*/       acc[0] += direction * forward[0] * 60;       acc[1] += direction * forward[1] * 60;       acc[2] += direction * forward[2] * 60;     }      /*if we sidestep add Left-Right-Booster acceleration (40)*/     if ( ucmd->sidemove )
    {
      /*are we accelerating left or right*/
      direction = (ucmd->sidemove<0) ? -1.0 : 1.0;        /*add only to x and y acceleration*/       acc[0] += right[0] * direction * 40;       acc[1] += right[1] * direction * 40;     }      /*if we crouch or jump add Up-Down-Booster acceleration (30)*/     if ( ucmd->upmove )
      acc[2] += ucmd->upmove > 0 ? 30 : -30;
 
    /*now apply some friction dependent on velocity (higher velocity results
      in higher friction), without acceleration this will reduce the velocity
      to 0 in a few steps*/
    ent->velocity[0] += -(ent->velocity[0]/6.0);
    ent->velocity[1] += -(ent->velocity[1]/6.0);
    ent->velocity[2] += -(ent->velocity[2]/7.0);
 
    /*then accelerate with the calculated values. If the new acceleration for
      a direction is smaller than an earlier, the friction will reduce the speed
      in that direction to the new value in a few steps, so if youre flying
      curves or around corners youre floating a little bit in the old direction*/
    VectorAdd( ent->velocity, acc, ent->velocity );
 
    /*round velocitys (is this necessary?)*/
    ent->velocity[0] = (float)((int)(ent->velocity[0]*8))/8;
    ent->velocity[1] = (float)((int)(ent->velocity[1]*8))/8;
    ent->velocity[2] = (float)((int)(ent->velocity[2]*8))/8;
 
    /*Bound velocitys so that friction and acceleration dont need to be
      synced on maxvelocitys*/
    for ( i=0 ; i<2 ; i++) /*allow z-velocity to be greater*/     {       if (ent->velocity[i] > 300)
        ent->velocity[i] = 300;
      else if (ent->velocity[i] < -300)
        ent->velocity[i] = -300;
    }
 
    /*add some gentle up and down when idle (not accelerating)*/
    if( VectorLength(acc) == 0 )
      Jet_ApplyLifting( ent );
 
  }//if ( ent->client->Jet_next_think...
 
  /*add rolling when we fly curves or boost left/right*/
  Jet_ApplyRolling( ent, right );

  if (ent->client->resp.class == 8)
  {
	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_AUTO, gi.soundindex("flyer/Flyidle1.wav"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+1.29;
	}
  }
  else if (ent->client->resp.class == 15)
  {
	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_AUTO, gi.soundindex("hover/hovidle1.wav"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+1.6;
	}
  }
  else if (ent->client->resp.class == 11)
  {
	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_AUTO, gi.soundindex("bosshovr/BHVENGN1.WAV"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+1.6;
	}
  } 
  else if (ent->client->resp.class == 16)
  {
	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_AUTO, gi.soundindex("floater/Fltsrch1.wav"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+2.4;
	}
  } 
}
 
/*end of jet.c*/


void ApplySupertankSound ( edict_t *ent )
{
	gi.sound (ent, CHAN_AUTO, gi.soundindex ("bosstank/btkengn1.wav"), 1, ATTN_NORM, 0);
	ent->client->resp.next_thrust_sound=level.time+7.1;
}

/*
=================
ApplyThrust

MUCE:
To add thrusting velocity to player
=================
*/

void ApplyThrust (edict_t *ent)
{

	vec3_t forward, right;
	vec3_t pack_pos, jet_vector;

	//MUCE:  add thrust to character

	if (ent->velocity[2] < -500)
		ent->velocity[2]+=((ent->velocity[2])/(-5));
	else if (ent->velocity[2] < 0)
		ent->velocity[2] += 100; 
	else
		ent->velocity[2]+=((1000-ent->velocity[2])/8);

	//MUCE:  add sparks

	AngleVectors(ent->client->v_angle, forward, right, NULL);
	VectorScale (forward, -7, pack_pos);
	VectorAdd (pack_pos, ent->s.origin, pack_pos);
	pack_pos[2] += (ent->viewheight);

	VectorScale (forward, -50, jet_vector);

	gi.WriteByte (svc_temp_entity);
	gi.WriteByte (TE_SPARKS);
	gi.WritePosition (pack_pos);
	gi.WriteDir (jet_vector);
	gi.multicast (pack_pos, MULTICAST_PVS);

	//MUCE: add sound 

	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_BODY, gi.soundindex("flyer/Flyidle1.wav"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+1.3;
	}
}

static void SP_FixCoopSpots (edict_t *self)
{
	edict_t	*spot;
	vec3_t	d;

	spot = NULL;

	while(1)
	{
		spot = G_Find(spot, FOFS(classname), "info_player_start");
		if (!spot)
			return;
		if (!spot->targetname)
			continue;
		VectorSubtract(self->s.origin, spot->s.origin, d);
		if (VectorLength(d) < 384)
		{
			if ((!self->targetname) || Q_stricmp(self->targetname, spot->targetname) != 0)
			{
//				gi.dprintf("FixCoopSpots changed %s at %s targetname from %s to %s\n", self->classname, vtos(self->s.origin), self->targetname, spot->targetname);
				self->targetname = spot->targetname;
			}
			return;
		}
	}
}

// now if that one wasn't ugly enough for you then try this one on for size
// some maps don't have any coop spots at all, so we need to create them
// where they should have been

static void SP_CreateCoopSpots (edict_t *self)
{
	edict_t	*spot;

	if(Q_stricmp(level.mapname, "security") == 0)
	{
		spot = G_Spawn();
		spot->classname = "info_player_coop";
		spot->s.origin[0] = 188 - 64;
		spot->s.origin[1] = -164;
		spot->s.origin[2] = 80;
		spot->targetname = "jail3";
		spot->s.angles[1] = 90;

		spot = G_Spawn();
		spot->classname = "info_player_coop";
		spot->s.origin[0] = 188 + 64;
		spot->s.origin[1] = -164;
		spot->s.origin[2] = 80;
		spot->targetname = "jail3";
		spot->s.angles[1] = 90;

		spot = G_Spawn();
		spot->classname = "info_player_coop";
		spot->s.origin[0] = 188 + 128;
		spot->s.origin[1] = -164;
		spot->s.origin[2] = 80;
		spot->targetname = "jail3";
		spot->s.angles[1] = 90;

		return;
	}
}


/*QUAKED info_player_start (1 0 0) (-16 -16 -24) (16 16 32)
The normal starting point for a level.
*/
void SP_info_player_start(edict_t *self)
{
	if (!coop->value)
		return;
	if(Q_stricmp(level.mapname, "security") == 0)
	{
		// invoke one of our gross, ugly, disgusting hacks
		self->think = SP_CreateCoopSpots;
		self->nextthink = level.time + FRAMETIME;
	}
}

/*QUAKED info_player_deathmatch (1 0 1) (-16 -16 -24) (16 16 32)
potential spawning position for deathmatch games
*/
void SP_info_player_deathmatch(edict_t *self)
{
	if (!deathmatch->value)
	{
		G_FreeEdict (self);
		return;
	}
	SP_misc_teleporter_dest (self);
}

/*QUAKED info_player_coop (1 0 1) (-16 -16 -24) (16 16 32)
potential spawning position for coop games
*/

void SP_info_player_coop(edict_t *self)
{
	if (!coop->value)
	{
		G_FreeEdict (self);
		return;
	}

	if((Q_stricmp(level.mapname, "jail2") == 0)   ||
	   (Q_stricmp(level.mapname, "jail4") == 0)   ||
	   (Q_stricmp(level.mapname, "mine1") == 0)   ||
	   (Q_stricmp(level.mapname, "mine2") == 0)   ||
	   (Q_stricmp(level.mapname, "mine3") == 0)   ||
	   (Q_stricmp(level.mapname, "mine4") == 0)   ||
	   (Q_stricmp(level.mapname, "lab") == 0)     ||
	   (Q_stricmp(level.mapname, "boss1") == 0)   ||
	   (Q_stricmp(level.mapname, "fact3") == 0)   ||
	   (Q_stricmp(level.mapname, "biggun") == 0)  ||
	   (Q_stricmp(level.mapname, "space") == 0)   ||
	   (Q_stricmp(level.mapname, "command") == 0) ||
	   (Q_stricmp(level.mapname, "power2") == 0) ||
	   (Q_stricmp(level.mapname, "strike") == 0))
	{
		// invoke one of our gross, ugly, disgusting hacks
		self->think = SP_FixCoopSpots;
		self->nextthink = level.time + FRAMETIME;
	}
}


/*QUAKED info_player_intermission (1 0 1) (-16 -16 -24) (16 16 32)
The deathmatch intermission point will be at one of these
Use 'angles' instead of 'angle', so you can set pitch or roll as well as yaw.  'pitch yaw roll'
*/
void SP_info_player_intermission(void)
{
}


//=======================================================================


void player_pain (edict_t *self, edict_t *other, float kick, int damage)
{
	// player pain is handled at the end of the frame in P_DamageFeedback
}


qboolean IsFemale (edict_t *ent)
{
	char		*info;

	if (!ent->client)
		return false;

	info = Info_ValueForKey (ent->client->pers.userinfo, "gender");
	if (info[0] == 'f' || info[0] == 'F')
		return true;
	return false;
}

qboolean IsNeutral (edict_t *ent)
{
	char		*info;

	if (!ent->client)
		return false;

	info = Info_ValueForKey (ent->client->pers.userinfo, "gender");
	if (info[0] != 'f' && info[0] != 'F' && info[0] != 'm' && info[0] != 'M')
		return true;
	return false;
}

void ClientObituary (edict_t *self, edict_t *inflictor, edict_t *attacker)
{
	int			mod;
	char		*message;
	char		*message2;
	qboolean	ff;

	if (coop->value && attacker->client)
		meansOfDeath |= MOD_FRIENDLY_FIRE;

	if (deathmatch->value || coop->value)
	{
		ff = meansOfDeath & MOD_FRIENDLY_FIRE;
		mod = meansOfDeath & ~MOD_FRIENDLY_FIRE;
		message = NULL;
		message2 = "";

		switch (mod)
		{
		case MOD_SUICIDE:
			message = "suicides";
			break;
		case MOD_FALLING:
			message = "cratered";
			break;
		case MOD_CRUSH:
			message = "was squished";
			break;
		case MOD_WATER:
			message = "sank like a rock";
			break;
		case MOD_SLIME:
			message = "melted";
			break;
		case MOD_LAVA:
			message = "does a back flip into the lava";
			break;
		case MOD_EXPLOSIVE:
		case MOD_BARREL:
			message = "blew up";
			break;
		case MOD_EXIT:
			message = "found a way out";
			break;
		case MOD_TARGET_LASER:
			message = "saw the light";
			break;
		case MOD_TARGET_BLASTER:
			message = "got blasted";
			break;
		case MOD_BOMB:
		case MOD_SPLASH:
		case MOD_TRIGGER_HURT:
			message = "was in the wrong place";
			break;
		}
		if (attacker == self)
		{
			switch (mod)
			{
			case MOD_HELD_GRENADE:
				message = "tried to put the pin back in";
				break;
			case MOD_HG_SPLASH:
			case MOD_G_SPLASH:
				if (IsNeutral(self))
					message = "tripped on its own grenade";
				else if (IsFemale(self))
					message = "tripped on her own grenade";
				else
					message = "tripped on his own grenade";
				break;
			case MOD_R_SPLASH:
				if (IsNeutral(self))
					message = "blew itself up";
				else if (IsFemale(self))
					message = "blew herself up";
				else
					message = "blew himself up";
				break;
			case MOD_BFG_BLAST:
				message = "should have used a smaller gun";
				break;
			default:
				if (IsNeutral(self))
					message = "killed itself";
				else if (IsFemale(self))
					message = "killed herself";
				else
					message = "killed himself";
				break;
			}
		}
		if (message)
		{
			gi.bprintf (PRINT_MEDIUM, "%s %s.\n", self->client->pers.netname, message);
			if (deathmatch->value)
				self->client->resp.score--;
			self->enemy = NULL;
			return;
		}

		self->enemy = attacker;
		if (attacker && attacker->client)
		{
			switch (mod)
			{
			case MOD_BLASTER:
				message = "was blasted by";
				break;
			case MOD_SHOTGUN:
				message = "was gunned down by";
				break;
			case MOD_SSHOTGUN:
				message = "was blown away by";
				message2 = "'s super shotgun";
				break;
			case MOD_MACHINEGUN:
				message = "was machinegunned by";
				break;
			case MOD_CHAINGUN:
				message = "was cut in half by";
				message2 = "'s chaingun";
				break;
			case MOD_GRENADE:
				message = "was popped by";
				message2 = "'s grenade";
				break;
			case MOD_G_SPLASH:
				message = "was shredded by";
				message2 = "'s shrapnel";
				break;
			case MOD_ROCKET:
				message = "ate";
				message2 = "'s rocket";
				break;
			case MOD_R_SPLASH:
				message = "almost dodged";
				message2 = "'s rocket";
				break;
			case MOD_HYPERBLASTER:
				message = "was melted by";
				message2 = "'s hyperblaster";
				break;
			case MOD_RAILGUN:
				message = "was railed by";
				break;
			case MOD_BFG_LASER:
				message = "saw the pretty lights from";
				message2 = "'s BFG";
				break;
			case MOD_BFG_BLAST:
				message = "was disintegrated by";
				message2 = "'s BFG blast";
				break;
			case MOD_BFG_EFFECT:
				message = "couldn't hide from";
				message2 = "'s BFG";
				break;
			case MOD_HANDGRENADE:
				message = "caught";
				message2 = "'s handgrenade";
				break;
			case MOD_HG_SPLASH:
				message = "didn't see";
				message2 = "'s handgrenade";
				break;
			case MOD_HELD_GRENADE:
				message = "feels";
				message2 = "'s pain";
				break;
			case MOD_TELEFRAG:
				message = "tried to invade";
				message2 = "'s personal space";
				break;
			case MOD_GLADRAIL:
				message = "was owned by";
				message2 = "'s shoulder railgun";
				break;
			case MOD_MEDICHYPER:
				message = "was melted to death by";
				message2 = "'s hyperblaster";
				break;
			case MOD_TANKBLAST:
				message = "took blaster bolts to the head from";
				message2 = "'s blaster";
				break;
			case MOD_SOLDBLAST:
				message = "was humiliated by";
				message2 = "'s blaster";
				break;
			case MOD_GUNNERCHAIN:
				message = "took bullets from";
				message2 = "'s armored chaingun";
				break;
			case MOD_INFMELEE:
				message = "was beaten to death from";
				message2 = "'s chaingun";
				break;
			case MOD_FLYSPIKES:
				message = "was slashed by";
				message2 = "'s blades";
				break;
			case MOD_GLADSPIKE:
				message = "was impaled by";
				message2 = "'s spike";
				break;
			case MOD_BRAINTENTACLE:
				message = "was eaten to death by";
				message2 = "'s live tentacles";
				break;
			case MOD_BRAINCLAWS:
				message = "took";
				message2 = "'s claws to the chest";
				break;
			case MOD_MUTSLASH:
				message = "has scars everywhere, thanks to";
				message2 = "'s claws";
				break;
			case MOD_BERCLUB:
				message = "was clubbed by";
				break;
			case MOD_BERSPIKE:
				message = "was spiked by";
				break;
			case MOD_IRONHAND:
				message = "was slapped and scarred from";
				message2 = "'s sharp hand";
				break;
			case MOD_STANKCHAIN:
				message = "took 12x12 bullets to the body from";
				message2 = "'s huge chaingun";
				break;
			case MOD_INFCHAIN:
				message = "was shot to death by";
				break;
			case MOD_BOSS2CH1:
				message = "took bullets from";
				message2 = "'s first chaingun";
				break;
			case MOD_BOSS2CH2:
				message = "took bullets from";
				message2 = "'s second chaingun";
				break;
			case MOD_TANKMACHINEGUN:
				message = "was pulverized from";
				message2 = "'s machinegun";
				break;
			case MOD_SOLDMACHINEGUN:
				message = "was shot by";
				message2 = "'s machinegun";
				break;
			case MOD_SOLDSHOTGUN:
				message = "was pelted from";
				message2 = "'s shotgun";
				break;
			}
			if (message)
			{
				gi.bprintf (PRINT_MEDIUM,"%s %s %s%s\n", self->client->pers.netname, message, attacker->client->pers.netname, message2);
				if (deathmatch->value)
				{
					if (ff)
						attacker->client->resp.score--;
					else
						attacker->client->resp.score++;
				}
				return;
			}
		}
	}

	gi.bprintf (PRINT_MEDIUM,"%s died.\n", self->client->pers.netname);
	if (deathmatch->value)
		self->client->resp.score--;
}


void Touch_Item (edict_t *ent, edict_t *other, cplane_t *plane, csurface_t *surf);

void TossClientWeapon (edict_t *self)
{
	gitem_t		*item;
	edict_t		*drop;
	qboolean	quad;
	float		spread;

	if (!deathmatch->value)
		return;

	item = self->client->pers.weapon;
	if (! self->client->pers.inventory[self->client->ammo_index] )
		item = NULL;
	if (item && (strcmp (item->pickup_name, "Soldier Blaster") == 0))
		item = NULL;

	if (!((int)(dmflags->value) & DF_QUAD_DROP))
		quad = false;
	else
		quad = (self->client->quad_framenum > (level.framenum + 10));

	if (item && quad)
		spread = 22.5;
	else
		spread = 0.0;

	if (item)
	{
		self->client->v_angle[YAW] -= spread;
		drop = Drop_Item (self, item);
		self->client->v_angle[YAW] += spread;
		drop->spawnflags = DROPPED_PLAYER_ITEM;
	}

	if (quad)
	{
		self->client->v_angle[YAW] += spread;
		drop = Drop_Item (self, FindItemByClassname ("item_quad"));
		self->client->v_angle[YAW] -= spread;
		drop->spawnflags |= DROPPED_PLAYER_ITEM;

		drop->touch = Touch_Item;
		drop->nextthink = level.time + (self->client->quad_framenum - level.framenum) * FRAMETIME;
		drop->think = G_FreeEdict;
	}
}


/*
==================
LookAtKiller
==================
*/
void LookAtKiller (edict_t *self, edict_t *inflictor, edict_t *attacker)
{
	vec3_t		dir;

	if (attacker && attacker != world && attacker != self)
	{
		VectorSubtract (attacker->s.origin, self->s.origin, dir);
	}
	else if (inflictor && inflictor != world && inflictor != self)
	{
		VectorSubtract (inflictor->s.origin, self->s.origin, dir);
	}
	else
	{
		self->client->killer_yaw = self->s.angles[YAW];
		return;
	}

	if (dir[0])
		self->client->killer_yaw = 180/M_PI*atan2(dir[1], dir[0]);
	else {
		self->client->killer_yaw = 0;
		if (dir[1] > 0)
			self->client->killer_yaw = 90;
		else if (dir[1] < 0)
			self->client->killer_yaw = -90;
	}
	if (self->client->killer_yaw < 0)
		self->client->killer_yaw += 360;
	

}


#define rndnum(y,z) ((random()*((z)-((y)+1)))+(y))
/*
==================
player_die
==================
*/
void player_die (edict_t *self, edict_t *inflictor, edict_t *attacker, int damage, vec3_t point)
{
	int		n;
	int		r;

	self->svflags &= ~SVF_NOCLIENT;

	VectorClear (self->avelocity);

	self->takedamage = DAMAGE_YES;
	self->movetype = MOVETYPE_TOSS;

	self->s.modelindex2 = 0;	// remove linked weapon model

	self->s.angles[0] = 0;
	self->s.angles[2] = 0;

	self->s.sound = 0;
	self->client->weapon_sound = 0;

	self->maxs[2] = -8;

//	self->solid = SOLID_NOT;
	self->svflags |= SVF_DEADMONSTER;

	if (!self->deadflag)
	{
		self->client->respawn_time = level.time + 1.0;
		LookAtKiller (self, inflictor, attacker);
		self->client->ps.pmove.pm_type = PM_DEAD;
		ClientObituary (self, inflictor, attacker);
		if (deathmatch->value)
			Cmd_Help_f (self);		// show scores

		// clear inventory
		// this is kind of ugly, but it's how we want to handle keys in coop
		for (n = 0; n < game.num_items; n++)
		{
			if (coop->value && itemlist[n].flags & IT_KEY)
				self->client->resp.coop_respawn.inventory[n] = self->client->pers.inventory[n];
			self->client->pers.inventory[n] = 0;
		}
	}

	// remove powerups
	self->client->quad_framenum = 0;
	self->client->invincible_framenum = 0;
	self->client->breather_framenum = 0;
	self->client->enviro_framenum = 0;
	self->flags &= ~FL_POWER_ARMOR;

	if (self->client->resp.class == 8)
	{
		BecomeExplosion1(self);
		gi.sound (self, CHAN_BODY, gi.soundindex ("flyer/flydeth1.wav"), 1, ATTN_NORM, 0);
		self->deadflag = DEAD_DEAD;
		gi.linkentity (self);
		ThrowClientHead (self, damage);

		self->takedamage = DAMAGE_NO;
		return;
	}
	if (self->client->resp.class == 16)
	{
		BecomeExplosion1(self);
		gi.sound (self, CHAN_BODY, gi.soundindex ("floater/fltdeth1.wav"), 1, ATTN_NORM, 0);
		self->deadflag = DEAD_DEAD;
		gi.linkentity (self);
		ThrowClientHead (self, damage);

		self->takedamage = DAMAGE_NO;
		return;
	}
	if (self->health < -40)
	{	// gib
		gi.sound (self, CHAN_BODY, gi.soundindex ("misc/udeath.wav"), 1, ATTN_NORM, 0);
		for (n= 0; n < 4; n++)
			ThrowGib (self, "models/objects/gibs/sm_meat/tris.md2", damage, GIB_ORGANIC);
		ThrowClientHead (self, damage);

		self->takedamage = DAMAGE_NO;
	}
	else
	{	// normal death
		if (!self->deadflag)
		{
			static int i;

			i = (i+1)%3;
			// start a death animation
			self->client->anim_priority = ANIM_DEATH;
			if (self->client->ps.pmove.pm_flags & PMF_DUCKED)
			{
				self->s.frame = FRAME_crdeath1-1;
				self->client->anim_end = FRAME_crdeath5;
			}
			else switch (i)
			{
			case 0:
				self->s.frame = FRAME_death101-1;
				self->client->anim_end = FRAME_death106;
				break;
			case 1:
				self->s.frame = FRAME_death201-1;
				self->client->anim_end = FRAME_death206;
				break;
			case 2:
				self->s.frame = FRAME_death301-1;
				self->client->anim_end = FRAME_death308;
				break;
			}
		if (self->client->resp.class == 1)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("soldier/soldeth1.wav, r"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 2)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("infantry/infdeth1.wav, r"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 3)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("tank/death.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 4)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("medic/meddeth1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 5)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("gunner/death1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 6)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("gladiator/glddeth2.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 7)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("bosstank/btkdeth1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 9)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("berserk/death.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 10)
		{
			r = rndnum (1, 3);
			gi.sound (self, CHAN_VOICE, gi.soundindex(va("chick/Chkdeth%d.wav", r)), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 11)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("bosshovr/BHVDETH1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 12)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("brain/brndeth1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 13)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("mutant/mutdeth1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 18)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("boss3/BS3DETH1.WAV"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 17)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("flipper/Flpdeth1.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 19)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("makron/death.wav"), 1, ATTN_NORM, 0);
		}
		else if (self->client->resp.class == 15)
		{
			gi.sound (self, CHAN_VOICE, gi.soundindex("hover/hovdeth1.wav"), 1, ATTN_NORM, 0);
		}
		}
	}

	self->deadflag = DEAD_DEAD;

	gi.linkentity (self);
}

//=======================================================================

/*
==============
InitClientPersistant

This is only called when the game first initializes in single player,
but is called after each death and level change in deathmatch
==============
*/
void InitClientPersistant (gclient_t *client)
{
    if (client->resp.class == 1) 
    {
        //Class 1
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 45;
	client->pers.max_health		= 45;
        item = FindItem("Shells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 25;

        item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 150;

        item = FindItem("Soldier Machinegun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        item = FindItem("Soldier Shotgun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        item = FindItem("Soldier Blaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 2)
    {
        //Class 2
        gitem_t         *item;


        memset (&client->pers, 0, sizeof(client->pers));


	client->pers.health			= 100;
	client->pers.max_health		= 100;
        item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 220;
        item = FindItem("Infantry Melee");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Infantry Chaingun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 3)
    {
        //Class 3
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 650;
	client->pers.max_health		= 650;

		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 50;
		item = FindItem("Rockets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 20;
		item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 200;
        item = FindItem("Tank Rocket Launcher");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Tank Machinegun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Tank Blaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 4)
    {
        //Class 4
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 300;
	client->pers.max_health		= 300;

		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 500;
        item = FindItem("Medic Hyperblaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 5)
    {
        //Class 5
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 175;
	client->pers.max_health		= 175;

		item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 300;
		item = FindItem("Grenades");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 40;
        item = FindItem("Gunner Grenade Launcher");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Gunner Chaingun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 6)
    {
        //Class 6
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 400;
	client->pers.max_health		= 400;

		item = FindItem("Slugs");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 100;
        item = FindItem("Gladiator Spike");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Gladiator Railgun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 7)
    {
        //Class 7
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));


	client->pers.health			= 900;
	client->pers.max_health		= 900;

		item = FindItem("Rockets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 35;
		item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 400;
        item = FindItem("SuperTank Rocket Launcher");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("SuperTank Chaingun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 8)
    {
        //Class 8
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));


	client->pers.health			= 70;
	client->pers.max_health		= 70;

		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 200;
        item = FindItem("Flyer Melee");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Flyer Hyperblaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 16)
    {
        //Class 16
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));


	client->pers.health			= 200;
	client->pers.max_health		= 200;

		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 350;
        item = FindItem("Floater Clamp");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1; 
		item = FindItem("Floater Melee");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Floater HyperBlaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 17)
    {
        //Class 17
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));


	client->pers.health			= 175;
	client->pers.max_health		= 175;

		item = FindItem("Shark Tail");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
		item = FindItem("Shark Bite");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 9)
    {
        //Class 9
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 240;
	client->pers.max_health		= 240;

		item = FindItem("Berserker Spike");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Berserker Club");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 10)
    {
        //Class 10
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 175;
	client->pers.max_health		= 175;

		item = FindItem("Rockets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 150;
        item = FindItem("Iron Maiden Slash");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
		item = FindItem("Iron Maiden Rocket Launcher");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 11)
    {
        //Class 11
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 1250;
	client->pers.max_health		= 1250;

		item = FindItem("Rockets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 75;
		item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 400;
        item = FindItem("Boss2 Rocket Launcher");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
		item = FindItem("Boss2 Chainguns");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 12)
    {
        //Class 12
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 550;
	client->pers.max_health		= 550;

        item = FindItem("Brain Claws");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
		item = FindItem("Brain Tentacles");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 13)
    {
        //Class 13
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 800;
	client->pers.max_health		= 800;

		item = FindItem("Mutant Claws");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 18)
    {
        //Class 18
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 2000;
	client->pers.max_health		= 2000;

		item = FindItem("Bullets");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 700;
        item = FindItem("Jorg Chainguns");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 19)
    {
        //Class 19
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 1700;
	client->pers.max_health		= 1700;


        item = FindItem("Makron Railgun");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
        item = FindItem("Makron BFG");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;
		item = FindItem("Slugs");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 230;
		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 450;
        item = FindItem("Makron Blaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else if (client->resp.class == 15)
    {
        //Class 15
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));

	client->pers.health			= 85;
	client->pers.max_health		= 85;

		item = FindItem("Cells");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 540;
        item = FindItem("Icarus HyperBlaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }
    else 
    {
        //Observer mode, doesn't really matter what they have
        gitem_t         *item;

        memset (&client->pers, 0, sizeof(client->pers));
    
	client->pers.health			= 1;
	client->pers.max_health		= 1;

        item = FindItem("Soldier Blaster");
        client->pers.selected_item = ITEM_INDEX(item);
        client->pers.inventory[client->pers.selected_item] = 1;

        client->pers.weapon = item;
    }

	client->pers.max_bullets	= 1000;
	client->pers.max_shells		= 1000;
	client->pers.max_rockets	= 1000;
	client->pers.max_grenades	= 1000;
	client->pers.max_cells		= 1000;
	client->pers.max_slugs		= 1000;

	client->pers.connected = true;
}


void InitClientResp (gclient_t *client)
{
	memset (&client->resp, 0, sizeof(client->resp));
	client->resp.enterframe = level.framenum;
	client->resp.coop_respawn = client->pers;
}

/*
==================
SaveClientData

Some information that should be persistant, like health, 
is still stored in the edict structure, so it needs to
be mirrored out to the client structure before all the
edicts are wiped.
==================
*/
void SaveClientData (void)
{
	int		i;
	edict_t	*ent;

	for (i=0 ; i<game.maxclients ; i++)
	{
		ent = &g_edicts[1+i];
		if (!ent->inuse)
			continue;
		game.clients[i].pers.health = ent->health;
		game.clients[i].pers.max_health = ent->max_health;
		game.clients[i].pers.savedFlags = (ent->flags & (FL_GODMODE|FL_NOTARGET|FL_POWER_ARMOR));
		if (coop->value)
			game.clients[i].pers.score = ent->client->resp.score;
	}
}

void FetchClientEntData (edict_t *ent)
{
	ent->health = ent->client->pers.health;
	ent->max_health = ent->client->pers.max_health;
	ent->flags |= ent->client->pers.savedFlags;
	if (coop->value)
		ent->client->resp.score = ent->client->pers.score;
}



/*
=======================================================================

  SelectSpawnPoint

=======================================================================
*/

/*
================
PlayersRangeFromSpot

Returns the distance to the nearest player from the given spot
================
*/
float	PlayersRangeFromSpot (edict_t *spot)
{
	edict_t	*player;
	float	bestplayerdistance;
	vec3_t	v;
	int		n;
	float	playerdistance;


	bestplayerdistance = 9999999;

	for (n = 1; n <= maxclients->value; n++)
	{
		player = &g_edicts[n];

		if (!player->inuse)
			continue;

		if (player->health <= 0)
			continue;

		VectorSubtract (spot->s.origin, player->s.origin, v);
		playerdistance = VectorLength (v);

		if (playerdistance < bestplayerdistance)
			bestplayerdistance = playerdistance;
	}

	return bestplayerdistance;
}

/*
================
SelectRandomDeathmatchSpawnPoint

go to a random point, but NOT the two points closest
to other players
================
*/
edict_t *SelectRandomDeathmatchSpawnPoint (void)
{
	edict_t	*spot, *spot1, *spot2;
	int		count = 0;
	int		selection;
	float	range, range1, range2;

	spot = NULL;
	range1 = range2 = 99999;
	spot1 = spot2 = NULL;

	while ((spot = G_Find (spot, FOFS(classname), "info_player_deathmatch")) != NULL)
	{
		count++;
		range = PlayersRangeFromSpot(spot);
		if (range < range1)
		{
			range1 = range;
			spot1 = spot;
		}
		else if (range < range2)
		{
			range2 = range;
			spot2 = spot;
		}
	}

	if (!count)
		return NULL;

	if (count <= 2)
	{
		spot1 = spot2 = NULL;
	}
	else
		count -= 2;

	selection = rand() % count;

	spot = NULL;
	do
	{
		spot = G_Find (spot, FOFS(classname), "info_player_deathmatch");
		if (spot == spot1 || spot == spot2)
			selection++;
	} while(selection--);

	return spot;
}

/*
================
SelectFarthestDeathmatchSpawnPoint

================
*/
edict_t *SelectFarthestDeathmatchSpawnPoint (void)
{
	edict_t	*bestspot;
	float	bestdistance, bestplayerdistance;
	edict_t	*spot;


	spot = NULL;
	bestspot = NULL;
	bestdistance = 0;
	while ((spot = G_Find (spot, FOFS(classname), "info_player_deathmatch")) != NULL)
	{
		bestplayerdistance = PlayersRangeFromSpot (spot);

		if (bestplayerdistance > bestdistance)
		{
			bestspot = spot;
			bestdistance = bestplayerdistance;
		}
	}

	if (bestspot)
	{
		return bestspot;
	}

	// if there is a player just spawned on each and every start spot
	// we have no choice to turn one into a telefrag meltdown
	spot = G_Find (NULL, FOFS(classname), "info_player_deathmatch");

	return spot;
}

edict_t *SelectDeathmatchSpawnPoint (void)
{
	if ( (int)(dmflags->value) & DF_SPAWN_FARTHEST)
		return SelectFarthestDeathmatchSpawnPoint ();
	else
		return SelectRandomDeathmatchSpawnPoint ();
}


edict_t *SelectCoopSpawnPoint (edict_t *ent)
{
	int		index;
	edict_t	*spot = NULL;
	char	*target;

	index = ent->client - game.clients;

	// player 0 starts in normal player spawn point
	if (!index)
		return NULL;

	spot = NULL;

	// assume there are four coop spots at each spawnpoint
	while (1)
	{
		spot = G_Find (spot, FOFS(classname), "info_player_coop");
		if (!spot)
			return NULL;	// we didn't have enough...

		target = spot->targetname;
		if (!target)
			target = "";
		if ( Q_stricmp(game.spawnpoint, target) == 0 )
		{	// this is a coop spawn point for one of the clients here
			index--;
			if (!index)
				return spot;		// this is it
		}
	}


	return spot;
}


/*
===========
SelectSpawnPoint

Chooses a player start, deathmatch start, coop start, etc
============
*/
void	SelectSpawnPoint (edict_t *ent, vec3_t origin, vec3_t angles)
{
	edict_t	*spot = NULL;

	if (deathmatch->value)
		spot = SelectDeathmatchSpawnPoint ();
	else if (coop->value)
		spot = SelectCoopSpawnPoint (ent);

	// find a single player start spot
	if (!spot)
	{
		while ((spot = G_Find (spot, FOFS(classname), "info_player_start")) != NULL)
		{
			if (!game.spawnpoint[0] && !spot->targetname)
				break;

			if (!game.spawnpoint[0] || !spot->targetname)
				continue;

			if (Q_stricmp(game.spawnpoint, spot->targetname) == 0)
				break;
		}

		if (!spot)
		{
			if (!game.spawnpoint[0])
			{	// there wasn't a spawnpoint without a target, so use any
				spot = G_Find (spot, FOFS(classname), "info_player_start");
			}
			if (!spot)
				gi.error ("Couldn't find spawn point %s\n", game.spawnpoint);
		}
	}

	VectorCopy (spot->s.origin, origin);
	origin[2] += 9;
	VectorCopy (spot->s.angles, angles);
}

//======================================================================


void InitBodyQue (void)
{
	int		i;
	edict_t	*ent;

	level.body_que = 0;
	for (i=0; i<BODY_QUEUE_SIZE ; i++)
	{
		ent = G_Spawn();
		ent->classname = "bodyque";
	}
}

void body_die (edict_t *self, edict_t *inflictor, edict_t *attacker, int damage, vec3_t point)
{
	int	n;

	if (self->health < -40)
	{
		gi.sound (self, CHAN_BODY, gi.soundindex ("misc/udeath.wav"), 1, ATTN_NORM, 0);
		for (n= 0; n < 4; n++)
			ThrowGib (self, "models/objects/gibs/sm_meat/tris.md2", damage, GIB_ORGANIC);
		self->s.origin[2] -= 48;
		ThrowClientHead (self, damage);
		self->takedamage = DAMAGE_NO;
	}
}

void CopyToBodyQue (edict_t *ent)
{
	edict_t		*body;

	// grab a body que and cycle to the next one
	body = &g_edicts[(int)maxclients->value + level.body_que + 1];
	level.body_que = (level.body_que + 1) % BODY_QUEUE_SIZE;

	// FIXME: send an effect on the removed body

	gi.unlinkentity (ent);

	gi.unlinkentity (body);
	body->s = ent->s;
	body->s.number = body - g_edicts;

	body->svflags = ent->svflags;
	VectorCopy (ent->mins, body->mins);
	VectorCopy (ent->maxs, body->maxs);
	VectorCopy (ent->absmin, body->absmin);
	VectorCopy (ent->absmax, body->absmax);
	VectorCopy (ent->size, body->size);
	body->solid = ent->solid;
	body->clipmask = ent->clipmask;
	body->owner = ent->owner;
	body->movetype = ent->movetype;

	body->die = body_die;
	body->takedamage = DAMAGE_YES;

	gi.linkentity (body);
}


void respawn (edict_t *self)
{
	if (deathmatch->value || coop->value)
	{
		// spectator's don't leave bodies
		if (self->movetype != MOVETYPE_NOCLIP)
			CopyToBodyQue (self);
		self->svflags &= ~SVF_NOCLIENT;
		PutClientInServer (self);

		// add a teleportation effect
		self->s.event = EV_PLAYER_TELEPORT;

		// hold in place briefly
		self->client->ps.pmove.pm_flags = PMF_TIME_TELEPORT;
		self->client->ps.pmove.pm_time = 14;

		self->client->respawn_time = level.time;

		return;
	}

	// restart the entire server
	gi.AddCommandString ("menu_loadgame\n");
}

/* 
 * only called when pers.spectator changes
 * note that resp.spectator should be the opposite of pers.spectator here
 */
void spectator_respawn (edict_t *ent)
{
	int i, numspec;

	// if the user wants to become a spectator, make sure he doesn't
	// exceed max_spectators

	if (ent->client->pers.spectator) {
		char *value = Info_ValueForKey (ent->client->pers.userinfo, "spectator");
		if (*spectator_password->string && 
			strcmp(spectator_password->string, "none") && 
			strcmp(spectator_password->string, value)) {
			gi.cprintf(ent, PRINT_HIGH, "Spectator password incorrect.\n");
			ent->client->pers.spectator = false;
			gi.WriteByte (svc_stufftext);
			gi.WriteString ("spectator 0\n");
			gi.unicast(ent, true);
			return;
		}

		// count spectators
		for (i = 1, numspec = 0; i <= maxclients->value; i++)
			if (g_edicts[i].inuse && g_edicts[i].client->pers.spectator)
				numspec++;

		if (numspec >= maxspectators->value) {
			gi.cprintf(ent, PRINT_HIGH, "Server spectator limit is full.");
			ent->client->pers.spectator = false;
			// reset his spectator var
			gi.WriteByte (svc_stufftext);
			gi.WriteString ("spectator 0\n");
			gi.unicast(ent, true);
			return;
		}
	} else {
		// he was a spectator and wants to join the game
		// he must have the right password
		char *value = Info_ValueForKey (ent->client->pers.userinfo, "password");
		if (*password->string && strcmp(password->string, "none") && 
			strcmp(password->string, value)) {
			gi.cprintf(ent, PRINT_HIGH, "Password incorrect.\n");
			ent->client->pers.spectator = true;
			gi.WriteByte (svc_stufftext);
			gi.WriteString ("spectator 1\n");
			gi.unicast(ent, true);
			return;
		}
	}

	// clear client on respawn
	ent->client->resp.score = ent->client->pers.score = 0;

	ent->svflags &= ~SVF_NOCLIENT;
	PutClientInServer (ent);

	// add a teleportation effect
	if (!ent->client->pers.spectator)  {
		// send effect
		gi.WriteByte (svc_muzzleflash);
		gi.WriteShort (ent-g_edicts);
		gi.WriteByte (MZ_LOGIN);
		gi.multicast (ent->s.origin, MULTICAST_PVS);

		// hold in place briefly
		ent->client->ps.pmove.pm_flags = PMF_TIME_TELEPORT;
		ent->client->ps.pmove.pm_time = 14;
	}

	ent->client->respawn_time = level.time;

	if (ent->client->pers.spectator) 
		gi.bprintf (PRINT_HIGH, "%s has moved to the sidelines\n", ent->client->pers.netname);
	else
		gi.bprintf (PRINT_HIGH, "%s joined the game\n", ent->client->pers.netname);
}

//==============================================================

void stuffcmd(edict_t *ent, char *s) 
{
gi.WriteByte (11); 
gi.WriteString (s);
gi.unicast (ent, true); 
} 

int AssignClassSpeed(edict_t *ent)
{
switch( ent->client->resp.class )
{
case 0: return 400;
break;
case 1: return 400;
break;
case 2: return 400;
break;
case 3: return 200;
break;
case 4: return 400;
break;
case 5: return 400;
break;
case 6: return 300;
break;
case 7: return 350;
break;
case 8: return 400;
break;
case 16: return 200;
break;
case 9: return 800;
break;
case 10: return 250;
break;
case 11: return 400;
break;
case 12: return 150;
break;
case 13: return 400;
break;
case 17: return 250;
break;
case 18: return 350;
break;
case 19: return 320;
break;
case 15: return 400;
break;
}
return ent->client->resp.class;
} 

void StuffSpeedByClass( edict_t *ent )
{
switch ( ent->client->resp.class )
{
case 0: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 1: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 2: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 3: stuffcmd( ent, "cl_forwardspeed 100\n" );
stuffcmd( ent, "cl_sidespeed 100\n" );
break;
case 4: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 5: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 6: stuffcmd( ent, "cl_forwardspeed 150\n" );
stuffcmd( ent, "cl_sidespeed 150\n" );
break;
case 7: stuffcmd( ent, "cl_forwardspeed 150\n" );
stuffcmd( ent, "cl_sidespeed 150\n" );
break;
case 8: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 16: stuffcmd( ent, "cl_forwardspeed 100\n" );
stuffcmd( ent, "cl_sidespeed 100\n" );
break;
case 9: stuffcmd( ent, "cl_forwardspeed 400\n" );
stuffcmd( ent, "cl_sidespeed 400\n" );
break;
case 10: stuffcmd( ent, "cl_forwardspeed 125\n" );
stuffcmd( ent, "cl_sidespeed 125\n" );
break;
case 17: stuffcmd( ent, "cl_forwardspeed 125\n" );
stuffcmd( ent, "cl_sidespeed 125\n" );
break;
case 11: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 150\n" );
break;
case 12: stuffcmd( ent, "cl_forwardspeed 75\n" );
stuffcmd( ent, "cl_sidespeed 75\n" );
break;
case 13: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
case 18: stuffcmd( ent, "cl_forwardspeed 150\n" );
stuffcmd( ent, "cl_sidespeed 150\n" );
break;
case 19: stuffcmd( ent, "cl_forwardspeed 160\n" );
stuffcmd( ent, "cl_sidespeed 160\n" );
break;
case 15: stuffcmd( ent, "cl_forwardspeed 200\n" );
stuffcmd( ent, "cl_sidespeed 200\n" );
break;
}
} 

/*
===========
PutClientInServer

Called when a player connects to a server or respawns in
a deathmatch.
============
*/
void PutClientInServer (edict_t *ent)
{
	vec3_t	mins = {-16, -16, -24};
	vec3_t	maxs = {16, 16, 32};
	int		index;
	vec3_t	spawn_origin, spawn_angles;
	gclient_t	*client;
	int		i;
	client_persistant_t	saved;
	client_respawn_t	resp;

	// find a spawn point
	// do it before setting health back up, so farthest
	// ranging doesn't count this client
	if (ent->client->resp.class == 17)
	{
/*--------------------------------*/
/*     new code starts here       */
/*--------------------------------*/
        VectorCopy (mins, ent->mins);
        VectorCopy (maxs, ent->maxs);
        if (findspawnpoint(ent)) // new line
        {
               VectorCopy(ent->s.origin, spawn_origin);
               VectorClear (spawn_angles);
               spawn_angles[YAW] = rand() % 360 - 180; // face a random direction
        }     
        else // couldn't find a good spot, so...
               SelectSpawnPoint (ent, spawn_origin, spawn_angles);
	}
/*--------------------------------*/
/*     new code ends here         */
/*--------------------------------*/
	else
		SelectSpawnPoint (ent, spawn_origin, spawn_angles);

	index = ent-g_edicts-1;
	client = ent->client;

	// deathmatch wipes most client data every spawn
	if (deathmatch->value)
	{
		char		userinfo[MAX_INFO_STRING];

		resp = client->resp;
		memcpy (userinfo, client->pers.userinfo, sizeof(userinfo));
		InitClientPersistant (client);
		ClientUserinfoChanged (ent, userinfo);
	}
	else if (coop->value)
	{
//		int			n;
		char		userinfo[MAX_INFO_STRING];

		resp = client->resp;
		memcpy (userinfo, client->pers.userinfo, sizeof(userinfo));
		// this is kind of ugly, but it's how we want to handle keys in coop
//		for (n = 0; n < game.num_items; n++)
//		{
//			if (itemlist[n].flags & IT_KEY)
//				resp.coop_respawn.inventory[n] = client->pers.inventory[n];
//		}
		resp.coop_respawn.game_helpchanged = client->pers.game_helpchanged;
		resp.coop_respawn.helpchanged = client->pers.helpchanged;
		client->pers = resp.coop_respawn;
		ClientUserinfoChanged (ent, userinfo);
		if (resp.score > client->pers.score)
			client->pers.score = resp.score;
	}
	else
	{
		memset (&resp, 0, sizeof(resp));
	}

	// clear everything but the persistant data
	saved = client->pers;
	memset (client, 0, sizeof(*client));
	client->pers = saved;
	if (client->pers.health <= 0)
		InitClientPersistant(client);
	client->resp = resp;

	// copy some data from the client to the entity
	FetchClientEntData (ent);

	// clear entity values
	ent->groundentity = NULL;
	ent->client = &game.clients[index];
	ent->takedamage = DAMAGE_AIM;
	ent->movetype = MOVETYPE_WALK;
	ent->viewheight = 22;
	ent->inuse = true;
	ent->classname = "player";
	ent->mass = 200;
	ent->solid = SOLID_BBOX;
	ent->deadflag = DEAD_NO;
	ent->air_finished = level.time + 12;
	ent->clipmask = MASK_PLAYERSOLID;
	ent->model = "players/male/tris.md2";
	ent->pain = player_pain;
	ent->die = player_die;
	ent->waterlevel = 0;
	ent->watertype = 0;
	ent->flags &= ~FL_NO_KNOCKBACK;
	ent->svflags &= ~SVF_DEADMONSTER;
	ent->client->resp.nextfootstep=0;

	StuffSpeedByClass( ent );
	ent->client->resp.class_speed = AssignClassSpeed( ent );

	if (deathmatch->value)
	{
		if (ent->client->resp.class <= 0) {
			ent->client->resp.class = 0;
			OpenMainMenus (ent); }
	}

	VectorCopy (mins, ent->mins);
	VectorCopy (maxs, ent->maxs);
	VectorClear (ent->velocity);

	// clear playerstate values
	memset (&ent->client->ps, 0, sizeof(client->ps));

	client->ps.pmove.origin[0] = spawn_origin[0]*8;
	client->ps.pmove.origin[1] = spawn_origin[1]*8;
	client->ps.pmove.origin[2] = spawn_origin[2]*8;

	if (deathmatch->value && ((int)dmflags->value & DF_FIXED_FOV))
	{
		client->ps.fov = 90;
	}
	else
	{
		client->ps.fov = atoi(Info_ValueForKey(client->pers.userinfo, "fov"));
		if (client->ps.fov < 1)
			client->ps.fov = 90;
		else if (client->ps.fov > 160)
			client->ps.fov = 160;
	}

	client->ps.gunindex = gi.modelindex(client->pers.weapon->view_model);

	// clear entity state values
	ent->s.effects = 0;
	ent->s.modelindex = 255;		// will use the skin specified model
	ent->s.modelindex2 = 255;		// custom gun model
	// sknum is player num and weapon number
	// weapon number will be added in changeweapon
	ent->s.skinnum = ent - g_edicts - 1;

	ent->s.frame = 0;
	VectorCopy (spawn_origin, ent->s.origin);
	ent->s.origin[2] += 1;	// make sure off ground
	VectorCopy (ent->s.origin, ent->s.old_origin);

	// set the delta angle
	for (i=0 ; i<3 ; i++)
	{
		client->ps.pmove.delta_angles[i] = ANGLE2SHORT(spawn_angles[i] - client->resp.cmd_angles[i]);
	}

	ent->s.angles[PITCH] = 0;
	ent->s.angles[YAW] = spawn_angles[YAW];
	ent->s.angles[ROLL] = 0;
	VectorCopy (ent->s.angles, client->ps.viewangles);
	VectorCopy (ent->s.angles, client->v_angle);

	// spawn a spectator
	if (client->pers.spectator) {
		client->chase_target = NULL;

		client->resp.spectator = true;

		ent->movetype = MOVETYPE_NOCLIP;
		ent->solid = SOLID_NOT;
		ent->svflags |= SVF_NOCLIENT;
		ent->client->ps.gunindex = 0;
		gi.linkentity (ent);
		return;
	} else
		client->resp.spectator = false;

	if (!KillBox (ent))
	{	// could't spawn in?
	}

	gi.linkentity (ent);

	// force the current weapon up
	client->newweapon = client->pers.weapon;
	ChangeWeapon (ent);
}

void EndObserverMode_Walking(edict_t* ent) 
{ 
    ent->movetype &= ~MOVETYPE_NOCLIP; 
    ent->solid &= ~SOLID_NOT; 
    ent->svflags &= ~SVF_NOCLIENT; 

    PutClientInServer (ent);

    if (level.intermissiontime)
    {
        MoveClientToIntermission (ent);
    }
    else
    {
        // send effect
        gi.WriteByte (svc_muzzleflash);
        gi.WriteShort (ent-g_edicts);
        gi.WriteByte (MZ_LOGIN);
        gi.multicast (ent->s.origin, MULTICAST_PVS);
    }

		ent->client->resp.thrusting = 0;
		ent->client->resp.next_thrust_sound=0;
		ent->client->resp.tread = 0;

    gi.bprintf (PRINT_HIGH, "%s has entered!\n", ent->client->pers.netname); 

} 

void EndObserverMode_Tread(edict_t* ent) 
{ 
    ent->movetype &= ~MOVETYPE_NOCLIP; 
    ent->solid &= ~SOLID_NOT; 
    ent->svflags &= ~SVF_NOCLIENT; 

    PutClientInServer (ent);

    if (level.intermissiontime)
    {
        MoveClientToIntermission (ent);
    }
    else
    {
        // send effect
        gi.WriteByte (svc_muzzleflash);
        gi.WriteShort (ent-g_edicts);
        gi.WriteByte (MZ_LOGIN);
        gi.multicast (ent->s.origin, MULTICAST_PVS);
    }

		ent->client->resp.thrusting = 0;
		ent->client->resp.next_thrust_sound=0;
		ent->client->resp.tread = 1;

    gi.bprintf (PRINT_HIGH, "%s has entered!\n", ent->client->pers.netname); 

} 


void EndObserverMode_Flying(edict_t* ent) 
{ 
    ent->movetype &= ~MOVETYPE_NOCLIP; 
    ent->solid &= ~SOLID_NOT; 
    ent->svflags &= ~SVF_NOCLIENT; 

    PutClientInServer (ent);

    if (level.intermissiontime)
    {
        MoveClientToIntermission (ent);
    }
    else
    {
        // send effect
        gi.WriteByte (svc_muzzleflash);
        gi.WriteShort (ent-g_edicts);
        gi.WriteByte (MZ_LOGIN);
        gi.multicast (ent->s.origin, MULTICAST_PVS);
    }

		ent->client->resp.thrusting=1;
		ent->client->resp.next_thrust_sound=0;
		ent->client->resp.tread = 0;

    gi.bprintf (PRINT_HIGH, "%s has entered!\n", ent->client->pers.netname); 
} 



/*
=====================
ClientBeginDeathmatch

A client has just connected to the server in 
deathmatch mode, so clear everything out before starting them.
=====================
*/
void ClientBeginDeathmatch (edict_t *ent)
{
	G_InitEdict (ent);

	InitClientResp (ent->client);

	// locate ent at a spawn point
	PutClientInServer (ent);

    ent->client->ps.gunindex = 0; 
    gi.linkentity (ent); 

	gi.bprintf (PRINT_HIGH, "%s entered the game\n", ent->client->pers.netname);

	// make sure all view stuff is valid
	ClientEndServerFrame (ent);
}


/*
===========
ClientBegin

called when a client has finished connecting, and is ready
to be placed into the game.  This will happen every level load.
============
*/
void ClientBegin (edict_t *ent)
{
	int		i;

	ent->client = game.clients + (ent - g_edicts - 1);

	if (deathmatch->value)
	{
		ClientBeginDeathmatch (ent);
		return;
	}

	// if there is already a body waiting for us (a loadgame), just
	// take it, otherwise spawn one from scratch
	if (ent->inuse == true)
	{
		// the client has cleared the client side viewangles upon
		// connecting to the server, which is different than the
		// state when the game is saved, so we need to compensate
		// with deltaangles
		for (i=0 ; i<3 ; i++)
			ent->client->ps.pmove.delta_angles[i] = ANGLE2SHORT(ent->client->ps.viewangles[i]);
	}
	else
	{
		// a spawn point will completely reinitialize the entity
		// except for the persistant data that was initialized at
		// ClientConnect() time
		G_InitEdict (ent);
		ent->classname = "player";
		InitClientResp (ent->client);
		PutClientInServer (ent);
	}

	if (level.intermissiontime)
	{
		MoveClientToIntermission (ent);
	}
	else
	{
		// send effect if in a multiplayer game
		if (game.maxclients > 1)
		{
			gi.WriteByte (svc_muzzleflash);
			gi.WriteShort (ent-g_edicts);
			gi.WriteByte (MZ_LOGIN);
			gi.multicast (ent->s.origin, MULTICAST_PVS);

			gi.bprintf (PRINT_HIGH, "%s entered the game\n", ent->client->pers.netname);
		}
	}

	// make sure all view stuff is valid
	ClientEndServerFrame (ent);
}

/*
===========
ClientUserInfoChanged

called whenever the player updates a userinfo variable.

The game can override any of the settings in place
(forcing skins or names, etc) before copying it off.
============
*/
void ClientUserinfoChanged (edict_t *ent, char *userinfo)
{
	char	*s;
	int		playernum;

	// check for malformed or illegal info strings
	if (!Info_Validate(userinfo))
	{
		strcpy (userinfo, "\\name\\badinfo\\skin\\male/grunt");
	}

	// set name
	s = Info_ValueForKey (userinfo, "name");
	strncpy (ent->client->pers.netname, s, sizeof(ent->client->pers.netname)-1);

	// set spectator
	s = Info_ValueForKey (userinfo, "spectator");
	// spectators are only supported in deathmatch
	if (deathmatch->value && *s && strcmp(s, "0"))
		ent->client->pers.spectator = true;
	else
		ent->client->pers.spectator = false;

	// set skin
	s = Info_ValueForKey (userinfo, "skin");

	playernum = ent-g_edicts-1;

	// combine name and skin into a configstring
	gi.configstring (CS_PLAYERSKINS+playernum, va("%s\\%s", ent->client->pers.netname, s) );

	// fov
	if (deathmatch->value && ((int)dmflags->value & DF_FIXED_FOV))
	{
		ent->client->ps.fov = 90;
	}
	else
	{
		ent->client->ps.fov = atoi(Info_ValueForKey(userinfo, "fov"));
		if (ent->client->ps.fov < 1)
			ent->client->ps.fov = 90;
		else if (ent->client->ps.fov > 160)
			ent->client->ps.fov = 160;
	}

	// handedness
	s = Info_ValueForKey (userinfo, "hand");
	if (strlen(s))
	{
		ent->client->pers.hand = atoi(s);
	}

	// save off the userinfo in case we want to check something later
	strncpy (ent->client->pers.userinfo, userinfo, sizeof(ent->client->pers.userinfo)-1);
}


/*
===========
ClientConnect

Called when a player begins connecting to the server.
The game can refuse entrance to a client by returning false.
If the client is allowed, the connection process will continue
and eventually get to ClientBegin()
Changing levels will NOT cause this to be called again, but
loadgames will.
============
*/
qboolean ClientConnect (edict_t *ent, char *userinfo)
{
	char	*value;

	// check to see if they are on the banned IP list
	value = Info_ValueForKey (userinfo, "ip");
	if (SV_FilterPacket(value)) {
		Info_SetValueForKey(userinfo, "rejmsg", "Banned.");
		return false;
	}

	// check for a spectator
	value = Info_ValueForKey (userinfo, "spectator");
	if (deathmatch->value && *value && strcmp(value, "0")) {
		int i, numspec;

		if (*spectator_password->string && 
			strcmp(spectator_password->string, "none") && 
			strcmp(spectator_password->string, value)) {
			Info_SetValueForKey(userinfo, "rejmsg", "Spectator password required or incorrect.");
			return false;
		}

		// count spectators
		for (i = numspec = 0; i < maxclients->value; i++)
			if (g_edicts[i+1].inuse && g_edicts[i+1].client->pers.spectator)
				numspec++;

		if (numspec >= maxspectators->value) {
			Info_SetValueForKey(userinfo, "rejmsg", "Server spectator limit is full.");
			return false;
		}
	} else {
		// check for a password
		value = Info_ValueForKey (userinfo, "password");
		if (*password->string && strcmp(password->string, "none") && 
			strcmp(password->string, value)) {
			Info_SetValueForKey(userinfo, "rejmsg", "Password required or incorrect.");
			return false;
		}
	}


	// they can connect
	ent->client = game.clients + (ent - g_edicts - 1);

	// if there is already a body waiting for us (a loadgame), just
	// take it, otherwise spawn one from scratch
	if (ent->inuse == false)
	{
		// clear the respawning variables
		InitClientResp (ent->client);
		if (!game.autosaved || !ent->client->pers.weapon)
			InitClientPersistant (ent->client);
	}

	ClientUserinfoChanged (ent, userinfo);

	if (game.maxclients > 1)
		gi.dprintf ("%s connected\n", ent->client->pers.netname);

	ent->svflags = 0; // make sure we start with known default
	ent->client->pers.connected = true;
	return true;
}

/*
===========
ClientDisconnect

Called when a player drops from the server.
Will not be called between levels.
============
*/
void ClientDisconnect (edict_t *ent)
{
	int		playernum;

	if (!ent->client)
		return;

	gi.bprintf (PRINT_HIGH, "%s disconnected\n", ent->client->pers.netname);

	// send effect
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ_LOGOUT);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	gi.unlinkentity (ent);
	ent->s.modelindex = 0;
	ent->solid = SOLID_NOT;
	ent->inuse = false;
	ent->classname = "disconnected";
	ent->client->pers.connected = false;

	playernum = ent-g_edicts-1;
	gi.configstring (CS_PLAYERSKINS+playernum, "");
}


//==============================================================


edict_t	*pm_passent;

// pmove doesn't need to know about passent and contentmask
trace_t	PM_trace (vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end)
{
	if (pm_passent->health > 0)
		return gi.trace (start, mins, maxs, end, pm_passent, MASK_PLAYERSOLID);
	else
		return gi.trace (start, mins, maxs, end, pm_passent, MASK_DEADSOLID);
}

unsigned CheckBlock (void *b, int c)
{
	int	v,i;
	v = 0;
	for (i=0 ; i<c ; i++)
		v+= ((byte *)b)[i];
	return v;
}
void PrintPmove (pmove_t *pm)
{
	unsigned	c1, c2;

	c1 = CheckBlock (&pm->s, sizeof(pm->s));
	c2 = CheckBlock (&pm->cmd, sizeof(pm->cmd));
	Com_Printf ("sv %3i:%i %i\n", pm->cmd.impulse, c1, c2);
}

//========================================================
void Cmd_Catapult_f(edict_t *ent) {
vec3_t forward,up;

  // Not available to dead or respawning players!
  //if (!G_ClientInGame(ent)) return;

  // Is ent currently airborne?
  if (ent->groundentity==NULL) return;

  else {
    // Physically throw the player up and forward!
    AngleVectors(ent->s.angles, forward, NULL, up);
    VectorCopy(forward, ent->movedir);
    VectorClear(ent->velocity);
    VectorMA(ent->velocity,450,forward,ent->velocity); // forward
    VectorMA(ent->velocity,400,up,ent->velocity); }    // upward
}

void MutantJumpAttack (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 40 + 10 * random();
	int			kick = 40;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_mutant (ent, start, forward, 80, damage, kick, 1, MOD_BLASTER);

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

/*
==============
ClientThink

This will be called once for each client frame, which will
usually be a couple times for each server frame.
==============
*/
void ClientThink (edict_t *ent, usercmd_t *ucmd)
{
	gclient_t	*client;
	edict_t	*other;
	int		i, j;
	pmove_t	pm;
	int backspeed;
	int sidespeed;
	int walkspeed;

	level.current_entity = ent;
	client = ent->client;

	if (level.intermissiontime)
	{
		client->ps.pmove.pm_type = PM_FREEZE;
		// can exit intermission after five seconds
		if (level.time > level.intermissiontime + 5.0 
			&& (ucmd->buttons & BUTTON_ANY) )
			level.exitintermission = true;
		return;
	}

	pm_passent = ent;

	if (ent->client->chase_target) {

		client->resp.cmd_angles[0] = SHORT2ANGLE(ucmd->angles[0]);
		client->resp.cmd_angles[1] = SHORT2ANGLE(ucmd->angles[1]);
		client->resp.cmd_angles[2] = SHORT2ANGLE(ucmd->angles[2]);

	} else {

		// set up for pmove
		memset (&pm, 0, sizeof(pm));

// speed cap code 

if ( (ucmd->forwardmove > ent->client->resp.class_speed) || (ucmd->sidemove > ent->client->resp.class_speed) )
{
ucmd->forwardmove = ent->client->resp.class_speed;
ucmd->sidemove = ent->client->resp.class_speed;
StuffSpeedByClass( ent );
} 

walkspeed = ent->client->resp.class_speed / 2;
// negative speed (left/back movement) check
if (ucmd->forwardmove < 0)
{
backspeed = ucmd->forwardmove + ent->client->resp.class_speed; 

if ( backspeed != 0 )
{
if ( backspeed != walkspeed )
{
ucmd->forwardmove = 0 - ent->client->resp.class_speed;
StuffSpeedByClass( ent );
}
}
}

if (ucmd->sidemove < 0)
{
sidespeed = ucmd->sidemove + ent->client->resp.class_speed; 

if ( sidespeed != 0 )
{
if (sidespeed != walkspeed)
{
ucmd->sidemove = 0 - ent->client->resp.class_speed;
StuffSpeedByClass( ent );
}
}
} 



		if (ent->movetype == MOVETYPE_NOCLIP)
			client->ps.pmove.pm_type = PM_SPECTATOR;
		else if (ent->s.modelindex != 255)
			client->ps.pmove.pm_type = PM_GIB;
		else if (ent->deadflag)
			client->ps.pmove.pm_type = PM_DEAD;
		else
			client->ps.pmove.pm_type = PM_NORMAL;

		client->ps.pmove.gravity = sv_gravity->value;

  /*ATTILA begin*/
  if ( ent->client->resp.thrusting==1 )
	Jet_ApplyJet( ent, ucmd );

  if (ent->client->resp.tread == 1)
  {
	if (ent->client->resp.next_thrust_sound < level.time)
	{
		gi.sound (ent, CHAN_AUTO, gi.soundindex("bosstank/btkengn1.wav"), 1, ATTN_NORM, 0);
		ent->client->resp.next_thrust_sound=level.time+6.6;
	}
  }
  /*ATTILA end*/
  /*ATTILA begin*/
  if ( ent->client->tread==1 )
    ApplySupertankSound (ent);
  /*ATTILA end*/

  if (deathmatch->value)
  {
    if (ent->client->resp.class < 1)
    {
        ent->solid = SOLID_NOT;
        ent->movetype = MOVETYPE_NOCLIP;
        ent->svflags |= SVF_NOCLIENT;
    }
  }

		pm.s = client->ps.pmove;

		for (i=0 ; i<3 ; i++)
		{
			pm.s.origin[i] = ent->s.origin[i]*8;
			pm.s.velocity[i] = ent->velocity[i]*8;
		}

		if (memcmp(&client->old_pmove, &pm.s, sizeof(pm.s)))
		{
			pm.snapinitial = true;
	//		gi.dprintf ("pmove changed!\n");
		}

		pm.cmd = *ucmd;

		pm.trace = PM_trace;	// adds default parms
		pm.pointcontents = gi.pointcontents;

		// perform a pmove
		gi.Pmove (&pm);

		// save results of pmove
		client->ps.pmove = pm.s;
		client->old_pmove = pm.s;

		for (i=0 ; i<3 ; i++)
		{
			ent->s.origin[i] = pm.s.origin[i]*0.125;
  /*ATTILA begin*/
  if ( !ent->client->resp.thrusting==1 || (ent->client->resp.thrusting==1&&(fabs((float)pm.s.velocity[i]*0.125) < fabs(ent->velocity[i]))) )
  /*ATTILA end*/

			ent->velocity[i] = pm.s.velocity[i]*0.125;
		}

		VectorCopy (pm.mins, ent->mins);
		VectorCopy (pm.maxs, ent->maxs);

		client->resp.cmd_angles[0] = SHORT2ANGLE(ucmd->angles[0]);
		client->resp.cmd_angles[1] = SHORT2ANGLE(ucmd->angles[1]);
		client->resp.cmd_angles[2] = SHORT2ANGLE(ucmd->angles[2]);

  /*ATTILA begin*/
  if ( ent->client->resp.class == 7 )
    if( pm.groundentity )               /*are we on ground*/
      if ( Jet_AvoidGround(ent) )       /*then lift us if possible*/
        pm.groundentity = NULL;         /*now we are no longer on ground*/
  /*ATTILA end*/
  /*ATTILA begin*/
  else if ( ent->client->resp.class == 8 )
    if( pm.groundentity )               /*are we on ground*/
      if ( Jet_AvoidGround(ent) )       /*then lift us if possible*/
        pm.groundentity = NULL;         /*now we are no longer on ground*/
  /*ATTILA end*/
  /*ATTILA begin*/
  else if ( ent->client->resp.class == 15 )
    if( pm.groundentity )               /*are we on ground*/
      if ( Jet_AvoidGround(ent) )       /*then lift us if possible*/
        pm.groundentity = NULL;         /*now we are no longer on ground*/
  /*ATTILA end*/
  /*ATTILA begin*/
  else if ( ent->client->resp.class == 16 )
    if( pm.groundentity )               /*are we on ground*/
      if ( Jet_AvoidGround(ent) )       /*then lift us if possible*/
        pm.groundentity = NULL;         /*now we are no longer on ground*/
  /*ATTILA end*/
  /*ATTILA begin*/
  else if ( ent->client->resp.class == 11 )
    if( pm.groundentity )               /*are we on ground*/
      if ( Jet_AvoidGround(ent) )       /*then lift us if possible*/
        pm.groundentity = NULL;         /*now we are no longer on ground*/
  /*ATTILA end*/

		if (ent->groundentity)
			ent->client->resp.mutant_jumped = 0;

		if (ent->groundentity && !pm.groundentity && (pm.cmd.upmove >= 10) && (pm.waterlevel == 0))
		{
			//gi.sound(ent, CHAN_VOICE, gi.soundindex("*jump1.wav"), 1, ATTN_NORM, 0);
			//PlayerNoise(ent, ent->s.origin, PNOISE_SELF);
			if (ent->client->resp.class == 13)
			{
				Cmd_Catapult_f (ent);
				gi.sound(ent, CHAN_VOICE, gi.soundindex("mutant/Mutsght1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (!ent->groundentity)
		{
			if (ent->client->resp.class == 13)
			{
				MutantJumpAttack (ent);
			}
		}


		//
		// Class Specific -
		// Check for.. out of.. water?
		// Barracuda shark cannot stay on land for long!
		//
		if (ent->client->resp.class == 17)
		{
			if (ent->waterlevel <= 1)
			{
				if (ent->groundentity)
				{
					if (ent->client->resp.nextdrown < level.time)
					{
						ent->health -= 3;
						gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flipper/flppain2.wav"), 1, ATTN_NORM, 0);
						ent->client->resp.nextdrown = level.time + 2;
					}
				}
			}
		}

		ent->viewheight = pm.viewheight;
		ent->waterlevel = pm.waterlevel;
		ent->watertype = pm.watertype;
		ent->groundentity = pm.groundentity;
		if (pm.groundentity)
			ent->groundentity_linkcount = pm.groundentity->linkcount;

		if (ent->deadflag)
		{
			client->ps.viewangles[ROLL] = 40;
			client->ps.viewangles[PITCH] = -15;
			client->ps.viewangles[YAW] = client->killer_yaw;
		}
		else
		{
			VectorCopy (pm.viewangles, client->v_angle);
			VectorCopy (pm.viewangles, client->ps.viewangles);
		}

		gi.linkentity (ent);

		if (ent->movetype != MOVETYPE_NOCLIP)
			G_TouchTriggers (ent);

		// touch other objects
		for (i=0 ; i<pm.numtouch ; i++)
		{
			other = pm.touchents[i];
			for (j=0 ; j<i ; j++)
				if (pm.touchents[j] == other)
					break;
			if (j != i)
				continue;	// duplicated
			if (!other->touch)
				continue;
			other->touch (other, ent, NULL, NULL);
		}

	}

	client->oldbuttons = client->buttons;
	client->buttons = ucmd->buttons;
	client->latched_buttons |= client->buttons & ~client->oldbuttons;

	// save light level the player is standing on for
	// monster sighting AI
	ent->light_level = ucmd->lightlevel;

	// fire weapon from final position if needed
	if (client->latched_buttons & BUTTON_ATTACK)
	{
		if (client->resp.spectator) {

			client->latched_buttons = 0;

			if (client->chase_target) {
				client->chase_target = NULL;
				client->ps.pmove.pm_flags &= ~PMF_NO_PREDICTION;
			} else
				GetChaseTarget(ent);

		} else if (!client->weapon_thunk) {
			client->weapon_thunk = true;
			Think_Weapon (ent);
		}
	}

	if (client->resp.spectator) {
		if (ucmd->upmove >= 10) {
			if (!(client->ps.pmove.pm_flags & PMF_JUMP_HELD)) {
				client->ps.pmove.pm_flags |= PMF_JUMP_HELD;
				if (client->chase_target)
					ChaseNext(ent);
				else
					GetChaseTarget(ent);
			}
		} else
			client->ps.pmove.pm_flags &= ~PMF_JUMP_HELD;
	}

	// update chase cam if being followed
	for (i = 1; i <= maxclients->value; i++) {
		other = g_edicts + i;
		if (other->inuse && other->client->chase_target == ent)
			UpdateChaseCam(other);
	}
}


/*
==============
ClientBeginServerFrame

This will be called once for each server frame, before running
any other entities in the world.
==============
*/
void ClientBeginServerFrame (edict_t *ent)
{
	gclient_t	*client;
	int			buttonMask;

	if (level.intermissiontime)
		return;

	client = ent->client;

	if (deathmatch->value &&
		client->pers.spectator != client->resp.spectator &&
		(level.time - client->respawn_time) >= 5) {
		spectator_respawn(ent);
		return;
	}

	// run weapon animations if it hasn't been done by a ucmd_t
	if (!client->weapon_thunk && !client->resp.spectator)
		Think_Weapon (ent);
	else
		client->weapon_thunk = false;

	if (ent->deadflag)
	{
		// wait for any button just going down
		if ( level.time > client->respawn_time)
		{
			// in deathmatch, only wait for attack button
			if (deathmatch->value)
				buttonMask = BUTTON_ATTACK;
			else
				buttonMask = -1;

			if ( ( client->latched_buttons & buttonMask ) ||
				(deathmatch->value && ((int)dmflags->value & DF_FORCE_RESPAWN) ) )
			{
				respawn(ent);
				client->latched_buttons = 0;
			}
		}
		return;
	}

	// add player trail so monsters can follow
	if (!deathmatch->value)
		if (!visible (ent, PlayerTrail_LastSpot() ) )
			PlayerTrail_Add (ent->s.old_origin);

	client->latched_buttons = 0;
}
