// g_weapon.c

#include "g_local.h"
#include "m_player.h"


static qboolean	is_quad;
static byte		is_silenced;

/*******mod_dist_point_to_line***********
* Simply calculates the distance from a point to a line
* point: the point for which we wish to calculate the distance
* linepoint: a point on the line
* linedir: the direction of the line
*****************************************/
float mod_dist_point_to_line(vec3_t point, vec3_t linepoint, vec3_t linedir)
{
        vec3_t temp, temp2;
 
        VectorSubtract(point, linepoint, temp);
        CrossProduct(temp, linedir, temp2);
 
        return VectorLength(temp2)/VectorLength(linedir);
}
 
/*******mod_GetLeadoffVec****************
* Author: Tim Matthews
* NOTE: This Code may be distributed and used freely as long as you give due credit to the author.
*
* Get's the direction in which to shoot at an enemy, taking into account his
* current velocity
*
*Arguments
*       self: you
*       fireorigin: the point from which the blast originates
*       rad: the radius of the bounding sphere. Only consider monsters inside this sphere
*       projspeed: the speed of the projectile being fired
*       bydist: true if judging best target by distance, otherwise test by angle
*       firedir: the direction we want to calculate
*************************************/
edict_t *mod_GetLeadoffVec(edict_t *self, vec3_t fireorigin, float rad, float projspeed, int bydist, vec3_t firedir)
{
        edict_t *other; /*any potential target besides yourself*/
        edict_t *best /*best victim so far*/;
 
        vec3_t viewvec;  /*your line of site*/
        vec3_t guessdir;
        vec3_t bestvec; /*the best guess for the direction of your blaster fire*/
        float bestdist, dist;  /*distance of the other guy and the shortest distance encountered*/
 
 
        vec3_t temp, temp2, otherdir;
        float d, t;
        double alpha, beta, rho;
        double a, b, c, t1, t2;
 
        //gi.dprintf("Checking Targets\n");
 
        AngleVectors (self->client->v_angle, viewvec, NULL, NULL); /*get the view direction*/
 
        best=NULL;
        other=findradius(NULL, fireorigin, rad); /*find something*/
 
        while (other) {
                 if ((other->svflags & SVF_MONSTER) && other->health>0 && other!=self) { /*might have to modify these*/
                         if (visible(self, other) && infront(self, other)) {
                                  /*player is in front and visible*/
 
                                  /*calculate lead off*/
                                  VectorSubtract(other->s.origin, other->s.old_origin, otherdir);
                                  alpha=VectorNormalize(otherdir); /*alpha = speed,  otherdir=direction vector*/
 
                                  if (alpha>.05 && projspeed>0) { /*if speed is significant, this value may have to be changed*/
                                          d=mod_dist_point_to_line(fireorigin, other->s.origin, otherdir); /*distance from the firepoint to the
                                                                                                                                                                                                                                                                                        the line on which the enemy is running*/
        
                                          beta=projspeed; /*our projectile speed*/
 
                                          VectorSubtract(fireorigin, other->s.origin, temp);
                                          CrossProduct(temp, otherdir, temp2);  //temp2 now holds the normal to a plane defined by fireorigin and other->s.origin
                                          CrossProduct(temp2, otherdir, temp);
                                          VectorNormalize(temp); /*temp holds the direction from the point to the line*/
 
                                          VectorScale(temp, d, guessdir);
                                          VectorAdd(guessdir, fireorigin, guessdir);
                                          VectorSubtract(guessdir, other->s.origin, guessdir);
                                          rho=VectorLength(guessdir); /*the length from other->s.origin to the point where the perpendicular vector from fireorigin intersects*/
 
                                          /*now, a little quadratic equation solving...*/
                                          a=alpha*alpha-beta*beta; 
                                          b=-2*alpha*rho;
                                          c=rho*rho+d*d;
 
                                          t1=(-b+sqrt(b*b-4*a*c))/(2*a);
                                          t2=(-b-sqrt(b*b-4*a*c))/(2*a);
 
                                          t=(t1>=0)?t1:t2; /*positive solution is the correct one*/
                                  
                                          VectorScale(otherdir, t*alpha, guessdir);
                                          VectorAdd(other->s.origin, guessdir, guessdir);
                                          VectorSubtract(guessdir, fireorigin, guessdir); /*now we have our best guess*/
                                  } else {
                                          /*enemy is standing still, so just get a simple direction vector*/
                                    VectorSubtract(other->s.origin, fireorigin, guessdir); 
                                  }
 
 
                                  if (bydist) {
                                          dist=guessdir[0]*guessdir[0]+guessdir[1]*guessdir[1]+guessdir[2]*guessdir[2];
 
                                          if (!best || dist<bestdist) {
                                                   best=other;
                                                   VectorCopy(guessdir, bestvec);
                                                   bestdist=dist;
                                          }
                                  } else {
                                          /*choose best as the person most in front of us*/
                                          VectorNormalize(guessdir);
                                          dist=DotProduct(viewvec, guessdir);
 
                                          if (!best || dist>bestdist) {
                                                   best=other;
                                                   VectorCopy(guessdir, bestvec);
                                                   bestdist=dist;
                                          }
                                 }
                         }
                 }
                 other=findradius(other, self->s.origin, rad); /*find the next entity*/
        }
        
        if (!best) /*No targets aquired, so just fire forward as usual*/
                 AngleVectors(self->client->v_angle, firedir, NULL, NULL);
        else {
                 //gi.dprintf("Target %s Aquired\n", best->classname);
                 VectorCopy(bestvec, firedir);
                 VectorNormalize(firedir);
        }
 
        return best;
}


void weapon_grenade_fire (edict_t *ent, qboolean held);

static void P_ProjectSource (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	if (client->pers.hand == LEFT_HANDED)
		_distance[1] *= -1;
	else if (client->pers.hand == CENTER_HANDED)
		_distance[1] = 0;
	G_ProjectSource (point, _distance, forward, right, result);
}

static void Left_ProjectSource (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	if (client->pers.hand == LEFT_HANDED)
		_distance[1] *= 8;
	else if (client->pers.hand == CENTER_HANDED)
		_distance[1] = 0;
	else if (client->pers.hand == RIGHT_HANDED)
		_distance[1] = -8;
	G_ProjectSource (point, _distance, forward, right, result);
}

void Left_Flyer_ProjectSource (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	if (client->pers.hand == LEFT_HANDED)
		_distance[1] *= 28;
	else if (client->pers.hand == CENTER_HANDED)
		_distance[1] = 0;
	else if (client->pers.hand == RIGHT_HANDED)
		_distance[1] = -28;
	G_ProjectSource (point, _distance, forward, right, result);
}

void Right_Flyer_ProjectSource (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	if (client->pers.hand == LEFT_HANDED)
		_distance[1] *= -28;
	else if (client->pers.hand == CENTER_HANDED)
		_distance[1] = 0;
	else if (client->pers.hand == RIGHT_HANDED)
		_distance[1] = 28;
	G_ProjectSource (point, _distance, forward, right, result);
}

void MBlaster_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("medic/medatck1.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Temp_Medic_Blaster_Fire (edict_t *ent)
{
	int		damage;

	damage = 2;

	MBlaster_Fire (ent, vec3_origin, damage, false, EF_BLASTER);
	ent->client->ps.gunframe++;
}

void MBlaster_FireRight (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	Right_Flyer_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);
	//ent->client->ps.gunframe++;

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flyer/flyatck3.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void MBlaster_FireLeft (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	Left_Flyer_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);
	//ent->client->ps.gunframe++;

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flyer/flyatck3.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Temp_Flyer_Blaster_FireRight (edict_t *ent)
{
	int		damage;

	damage = 2;

	MBlaster_FireRight (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
}

void Temp_Flyer_Blaster_FireLeft (edict_t *ent)
{
	int		damage;

	damage = 2;

	MBlaster_FireLeft (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
}

void GrenLaunch_Source (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	if (client->pers.hand == LEFT_HANDED)
		_distance[1] *= 5;
	else if (client->pers.hand == CENTER_HANDED)
		_distance[1] = 0;
	else if (client->pers.hand == RIGHT_HANDED)
		_distance[1] = -5;
	G_ProjectSource (point, _distance, forward, right, result);
}

void Middle_Source (gclient_t *client, vec3_t point, vec3_t distance, vec3_t forward, vec3_t right, vec3_t result)
{
	vec3_t	_distance;

	VectorCopy (distance, _distance);
	_distance[1] = 0;
	G_ProjectSource (point, _distance, forward, right, result);
}
/*
===============
PlayerNoise

Each player can have two noise objects associated with it:
a personal noise (jumping, pain, weapon firing), and a weapon
target noise (bullet wall impacts)

Monsters that don't directly see the player can move
to a noise in hopes of seeing the player from there.
===============
*/
void PlayerNoise(edict_t *who, vec3_t where, int type)
{
	edict_t		*noise;

	if (type == PNOISE_WEAPON)
	{
		if (who->client->silencer_shots)
		{
			who->client->silencer_shots--;
			return;
		}
	}

	if (deathmatch->value)
		return;

	if (who->flags & FL_NOTARGET)
		return;


	if (!who->mynoise)
	{
		noise = G_Spawn();
		noise->classname = "player_noise";
		VectorSet (noise->mins, -8, -8, -8);
		VectorSet (noise->maxs, 8, 8, 8);
		noise->owner = who;
		noise->svflags = SVF_NOCLIENT;
		who->mynoise = noise;

		noise = G_Spawn();
		noise->classname = "player_noise";
		VectorSet (noise->mins, -8, -8, -8);
		VectorSet (noise->maxs, 8, 8, 8);
		noise->owner = who;
		noise->svflags = SVF_NOCLIENT;
		who->mynoise2 = noise;
	}

	if (type == PNOISE_SELF || type == PNOISE_WEAPON)
	{
		noise = who->mynoise;
		level.sound_entity = noise;
		level.sound_entity_framenum = level.framenum;
	}
	else // type == PNOISE_IMPACT
	{
		noise = who->mynoise2;
		level.sound2_entity = noise;
		level.sound2_entity_framenum = level.framenum;
	}

	VectorCopy (where, noise->s.origin);
	VectorSubtract (where, noise->maxs, noise->absmin);
	VectorAdd (where, noise->maxs, noise->absmax);
	noise->teleport_time = level.time;
	gi.linkentity (noise);
}


qboolean Pickup_Weapon (edict_t *ent, edict_t *other)
{
	int			index;
	gitem_t		*ammo;

	index = ITEM_INDEX(ent->item);

	if ( ( ((int)(dmflags->value) & DF_WEAPONS_STAY) || coop->value) 
		&& other->client->pers.inventory[index])
	{
		if (!(ent->spawnflags & (DROPPED_ITEM | DROPPED_PLAYER_ITEM) ) )
			return false;	// leave the weapon for others to pickup
	}

	other->client->pers.inventory[index]++;

	if (!(ent->spawnflags & DROPPED_ITEM) )
	{
		// give them some ammo with it
		ammo = FindItem (ent->item->ammo);
		if ( (int)dmflags->value & DF_INFINITE_AMMO )
			Add_Ammo (other, ammo, 1000);
		else
			Add_Ammo (other, ammo, ammo->quantity);

		if (! (ent->spawnflags & DROPPED_PLAYER_ITEM) )
		{
			if (deathmatch->value)
			{
				if ((int)(dmflags->value) & DF_WEAPONS_STAY)
					ent->flags |= FL_RESPAWN;
				else
					SetRespawn (ent, 30);
			}
			if (coop->value)
				ent->flags |= FL_RESPAWN;
		}
	}

	if (other->client->pers.weapon != ent->item && 
		(other->client->pers.inventory[index] == 1) &&
		( !deathmatch->value || other->client->pers.weapon == FindItem("Soldier blaster") ) )
		other->client->newweapon = ent->item;

	return true;
}


/*
===============
ChangeWeapon

The old weapon has been dropped all the way, so make the new one
current
===============
*/
void ChangeWeapon (edict_t *ent)
{
	int i;

	if (ent->client->grenade_time)
	{
		ent->client->grenade_time = level.time;
		ent->client->weapon_sound = 0;
		weapon_grenade_fire (ent, false);
		ent->client->grenade_time = 0;
	}

	ent->client->pers.lastweapon = ent->client->pers.weapon;
	ent->client->pers.weapon = ent->client->newweapon;
	ent->client->newweapon = NULL;
	ent->client->machinegun_shots = 0;

	// set visible model
	if (ent->s.modelindex == 255) {
		if (ent->client->pers.weapon)
			i = ((ent->client->pers.weapon->weapmodel & 0xff) << 8);
		else
			i = 0;
		ent->s.skinnum = (ent - g_edicts - 1) | i;
	}

	if (ent->client->pers.weapon && ent->client->pers.weapon->ammo)
		ent->client->ammo_index = ITEM_INDEX(FindItem(ent->client->pers.weapon->ammo));
	else
		ent->client->ammo_index = 0;

	if (!ent->client->pers.weapon)
	{	// dead
		ent->client->ps.gunindex = 0;
		return;
	}

	ent->client->weaponstate = WEAPON_ACTIVATING;
	ent->client->ps.gunframe = 0;
	ent->client->ps.gunindex = gi.modelindex(ent->client->pers.weapon->view_model);

	ent->client->anim_priority = ANIM_PAIN;
	if(ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
			ent->s.frame = FRAME_crpain1;
			ent->client->anim_end = FRAME_crpain4;
	}
	else
	{
			ent->s.frame = FRAME_pain301;
			ent->client->anim_end = FRAME_pain304;
			
	}
}

/*
=================
NoAmmoWeaponChange
=================
*/
void NoAmmoWeaponChange (edict_t *ent)
{
	gi.cprintf (ent, PRINT_HIGH, "NoAmmoWeaponChange called, ignored...\n");
}

/*
=================
Think_Weapon

Called by ClientBeginServerFrame and ClientThink
=================
*/
void Think_Weapon (edict_t *ent)
{
	// if just died, put the weapon away
	if (ent->health < 1)
	{
		ent->client->newweapon = NULL;
		ChangeWeapon (ent);
	}

	// call active weapon think routine
	if (ent->client->pers.weapon && ent->client->pers.weapon->weaponthink)
	{
		is_quad = (ent->client->quad_framenum > level.framenum);
		if (ent->client->silencer_shots)
			is_silenced = MZ_SILENCED;
		else
			is_silenced = 0;
		ent->client->pers.weapon->weaponthink (ent);
	}
}


/*
================
Use_Weapon

Make the weapon ready if there is ammo
================
*/
void Use_Weapon (edict_t *ent, gitem_t *item)
{
	int			ammo_index;
	gitem_t		*ammo_item;

	// see if we're already using it
	if (item == ent->client->pers.weapon)
		return;

	// change to this weapon when down
	ent->client->newweapon = item;
}



/*
================
Drop_Weapon
================
*/
void Drop_Weapon (edict_t *ent, gitem_t *item)
{
	int		index;

	if ((int)(dmflags->value) & DF_WEAPONS_STAY)
		return;

	index = ITEM_INDEX(item);
	// see if we're already using it
	if ( ((item == ent->client->pers.weapon) || (item == ent->client->newweapon))&& (ent->client->pers.inventory[index] == 1) )
	{
		gi.cprintf (ent, PRINT_HIGH, "Can't drop current weapon\n");
		return;
	}

	Drop_Item (ent, item);
	ent->client->pers.inventory[index]--;
}


/*
================
Weapon_Generic

A generic function to handle the basics of weapon thinking
================
*/
#define FRAME_FIRE_FIRST		(FRAME_ACTIVATE_LAST + 1)
#define FRAME_IDLE_FIRST		(FRAME_FIRE_LAST + 1)
#define FRAME_DEACTIVATE_FIRST	(FRAME_IDLE_LAST + 1)

void Weapon_Generic (edict_t *ent, int FRAME_ACTIVATE_LAST, int FRAME_FIRE_LAST, int FRAME_IDLE_LAST, int FRAME_DEACTIVATE_LAST, int *pause_frames, int *fire_frames, void (*fire)(edict_t *ent))
{
	int		n;

	if(ent->deadflag || ent->s.modelindex != 255) // VWep animations screw up corpses
	{
		return;
	}

	if (ent->client->weaponstate == WEAPON_DROPPING)
	{
		if (ent->client->ps.gunframe == FRAME_DEACTIVATE_LAST)
		{
			ChangeWeapon (ent);
			return;
		}
		else if ((FRAME_DEACTIVATE_LAST - ent->client->ps.gunframe) == 4)
		{
			ent->client->anim_priority = ANIM_REVERSE;
			if(ent->client->ps.pmove.pm_flags & PMF_DUCKED)
			{
				ent->s.frame = FRAME_crpain4+1;
				ent->client->anim_end = FRAME_crpain1;
			}
			else
			{
				ent->s.frame = FRAME_pain304+1;
				ent->client->anim_end = FRAME_pain301;
				
			}
		}

		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->weaponstate == WEAPON_ACTIVATING)
	{
		if (ent->client->ps.gunframe == FRAME_ACTIVATE_LAST)
		{
			ent->client->weaponstate = WEAPON_READY;
			ent->client->ps.gunframe = FRAME_IDLE_FIRST;
			return;
		}

		ent->client->ps.gunframe++;
		return;
	}

	if ((ent->client->newweapon) && (ent->client->weaponstate != WEAPON_FIRING))
	{
		ent->client->weaponstate = WEAPON_DROPPING;
		ent->client->ps.gunframe = FRAME_DEACTIVATE_FIRST;

		if ((FRAME_DEACTIVATE_LAST - FRAME_DEACTIVATE_FIRST) < 4)
		{
			ent->client->anim_priority = ANIM_REVERSE;
			if(ent->client->ps.pmove.pm_flags & PMF_DUCKED)
			{
				ent->s.frame = FRAME_crpain4+1;
				ent->client->anim_end = FRAME_crpain1;
			}
			else
			{
				ent->s.frame = FRAME_pain304+1;
				ent->client->anim_end = FRAME_pain301;
				
			}
		}
		return;
	}

	if (ent->client->weaponstate == WEAPON_READY)
	{
		if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
		{
			ent->client->latched_buttons &= ~BUTTON_ATTACK;
			if ((!ent->client->ammo_index) || 
				( ent->client->pers.inventory[ent->client->ammo_index] >= ent->client->pers.weapon->quantity))
			{
				ent->client->ps.gunframe = FRAME_FIRE_FIRST;
				ent->client->weaponstate = WEAPON_FIRING;

				// start the animation
				ent->client->anim_priority = ANIM_ATTACK;
				if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
				{
					ent->s.frame = FRAME_crattak1-1;
					ent->client->anim_end = FRAME_crattak9;
				}
				else
				{
					ent->s.frame = FRAME_attack1-1;
					ent->client->anim_end = FRAME_attack8;
				}
			}
			else
			{
				if (level.time >= ent->pain_debounce_time)
				{
					gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
					ent->pain_debounce_time = level.time + 1;
				}
			}
		}
		else
		{
			if (ent->client->ps.gunframe == FRAME_IDLE_LAST)
			{
				ent->client->ps.gunframe = FRAME_IDLE_FIRST;
				return;
			}

			if (pause_frames)
			{
				for (n = 0; pause_frames[n]; n++)
				{
					if (ent->client->ps.gunframe == pause_frames[n])
					{
						if (rand()&15)
							return;
					}
				}
			}

			ent->client->ps.gunframe++;
			return;
		}
	}

	if (ent->client->weaponstate == WEAPON_FIRING)
	{
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Gunner Grenade Launcher") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 8)
					ent->client->ps.gunframe = 5;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Icarus HyperBlaster") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 5)
					ent->client->ps.gunframe = 2;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Shark Bite") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 21)
					ent->client->ps.gunframe = 10;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Shark Tail") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 20)
					ent->client->ps.gunframe = 9;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Tank Blaster") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 6)
					ent->client->ps.gunframe = 3;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Jorg Chainguns") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 6)
					ent->client->ps.gunframe = 2;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Tank Rocket Launcher") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 23)
					ent->client->ps.gunframe = 20;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Mutant Claws") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 12)
					ent->client->ps.gunframe = 4;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Medic Hyperblaster") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 13)
					ent->client->ps.gunframe = 12;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Medic Hyperblaster") == 0)
		{
			if (ent->client->ps.gunframe == 4)
			{
				Temp_Medic_Blaster_Fire (ent);
			}
			if (ent->client->ps.gunframe == 8)
			{
				Temp_Medic_Blaster_Fire (ent);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Flyer Hyperblaster") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 5)
					ent->client->ps.gunframe = 2;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Flyer Hyperblaster") == 0)
		{
			if (ent->client->ps.gunframe == 2)
			{
				Temp_Flyer_Blaster_FireLeft (ent);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Flyer Hyperblaster") == 0)
		{
			if (ent->client->ps.gunframe == 2)
			{
				Temp_Flyer_Blaster_FireRight (ent);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Flyer Hyperblaster") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				Temp_Flyer_Blaster_FireLeft (ent);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Flyer Hyperblaster") == 0)
		{
			if (ent->client->ps.gunframe == 4)
			{
				Temp_Flyer_Blaster_FireRight (ent);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Infantry Melee") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 14)
					ent->client->ps.gunframe = 4;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Soldier Blaster") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 9)
					ent->client->ps.gunframe = 4;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Gladiator Spike") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 9)
					ent->client->ps.gunframe = 3;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Brain Tentacles") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 14)
					ent->client->ps.gunframe = 9;
			}
		}
		
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Brain Tentacles") == 0)
		{
			if (ent->client->ps.gunframe == 8)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("brain/Brnatck1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Makron Railgun") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("makron/rail_up.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Shark Bite") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flipper/flpatck1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Shark Tail") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flipper/flpatck2.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Brain Claws") == 0)
		{
			if (ent->client->ps.gunframe == 7)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("brain/melee1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Brain Claws") == 0)
		{
			if (ent->client->ps.gunframe == 12)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("brain/melee2.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Soldier Shotgun") == 0)
		{
			if (ent->client->ps.gunframe == 10)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("infantry/infatck3.wav"), 1, ATTN_NORM, 0);
			}
		}
		//
		//
		//	Iron Maiden
		//
		//
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Iron Maiden Rocket Launcher") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("chick/Chkatck1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Iron Maiden Rocket Launcher") == 0)
		{
			if (ent->client->ps.gunframe == 25)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("chick/Chkatck5.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Iron Maiden Rocket Launcher") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 30)
					ent->client->ps.gunframe = 17;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Iron Maiden Slash") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 11)
					ent->client->ps.gunframe = 2;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Boss2 Chainguns") == 0)
		{
			if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
			{
				if (ent->client->ps.gunframe == 5)
					ent->client->ps.gunframe = 3;
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Infantry Melee") == 0)
		{
			if (ent->client->ps.gunframe == 8)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("infantry/infatck2.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Berserker Club") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("berserk/attack.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Berserker Spike") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("berserk/attack.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Gladiator Spike") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("gladiator/melee1.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Gladiator Railgun") == 0)
		{
			if (ent->client->ps.gunframe == 3)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("gladiator/railgun.wav"), 1, ATTN_NORM, 0);
			}
		}
		if (Q_stricmp(ent->client->pers.weapon->pickup_name, "Gunner Chaingun") == 0)
		{
			if (ent->client->ps.gunframe == 4)
			{
				gi.sound(ent, CHAN_WEAPON, gi.soundindex ("gunner/gunatck1.wav"), 1, ATTN_NORM, 0);
			}
		}
		for (n = 0; fire_frames[n]; n++)
		{
			if (ent->client->ps.gunframe == fire_frames[n])
			{
				if (ent->client->quad_framenum > level.framenum)
					gi.sound(ent, CHAN_ITEM, gi.soundindex("items/damage3.wav"), 1, ATTN_NORM, 0);

				fire (ent);
				break;
			}
		}

		if (!fire_frames[n])
			ent->client->ps.gunframe++;

		if (ent->client->ps.gunframe == FRAME_IDLE_FIRST+1)
			ent->client->weaponstate = WEAPON_READY;
	}
}


/*
======================================================================

GRENADE

======================================================================
*/

#define GRENADE_TIMER		3.0
#define GRENADE_MINSPEED	400
#define GRENADE_MAXSPEED	800

void weapon_grenade_fire (edict_t *ent, qboolean held)
{
	vec3_t	offset;
	vec3_t	forward, right;
	vec3_t	start;
	int		damage = 125;
	float	timer;
	int		speed;
	float	radius;

	radius = damage+40;
	if (is_quad)
		damage *= 4;

	VectorSet(offset, 8, 8, ent->viewheight-8);
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	timer = ent->client->grenade_time - level.time;
	speed = GRENADE_MINSPEED + (GRENADE_TIMER - timer) * ((GRENADE_MAXSPEED - GRENADE_MINSPEED) / GRENADE_TIMER);
	fire_grenade2 (ent, start, forward, damage, speed, timer, radius, held);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->grenade_time = level.time + 1.0;

	if(ent->deadflag || ent->s.modelindex != 255) // VWep animations screw up corpses
	{
		return;
	}

	if (ent->health <= 0)
		return;

	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->client->anim_priority = ANIM_ATTACK;
		ent->s.frame = FRAME_crattak1-1;
		ent->client->anim_end = FRAME_crattak3;
	}
	else
	{
		ent->client->anim_priority = ANIM_REVERSE;
		ent->s.frame = FRAME_wave08;
		ent->client->anim_end = FRAME_wave01;
	}
}

void Weapon_Grenade (edict_t *ent)
{
	if ((ent->client->newweapon) && (ent->client->weaponstate == WEAPON_READY))
	{
		ChangeWeapon (ent);
		return;
	}

	if (ent->client->weaponstate == WEAPON_ACTIVATING)
	{
		ent->client->weaponstate = WEAPON_READY;
		ent->client->ps.gunframe = 16;
		return;
	}

	if (ent->client->weaponstate == WEAPON_READY)
	{
		if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
		{
			ent->client->latched_buttons &= ~BUTTON_ATTACK;
			if (ent->client->pers.inventory[ent->client->ammo_index])
			{
				ent->client->ps.gunframe = 1;
				ent->client->weaponstate = WEAPON_FIRING;
				ent->client->grenade_time = 0;
			}
			else
			{
				if (level.time >= ent->pain_debounce_time)
				{
					gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
					ent->pain_debounce_time = level.time + 1;
				}
			}
			return;
		}

		if ((ent->client->ps.gunframe == 29) || (ent->client->ps.gunframe == 34) || (ent->client->ps.gunframe == 39) || (ent->client->ps.gunframe == 48))
		{
			if (rand()&15)
				return;
		}

		if (++ent->client->ps.gunframe > 48)
			ent->client->ps.gunframe = 16;
		return;
	}

	if (ent->client->weaponstate == WEAPON_FIRING)
	{
		if (ent->client->ps.gunframe == 5)
			gi.sound(ent, CHAN_WEAPON, gi.soundindex("weapons/hgrena1b.wav"), 1, ATTN_NORM, 0);

		if (ent->client->ps.gunframe == 11)
		{
			if (!ent->client->grenade_time)
			{
				ent->client->grenade_time = level.time + GRENADE_TIMER + 0.2;
				ent->client->weapon_sound = gi.soundindex("weapons/hgrenc1b.wav");
			}

			// they waited too long, detonate it in their hand
			if (!ent->client->grenade_blew_up && level.time >= ent->client->grenade_time)
			{
				ent->client->weapon_sound = 0;
				weapon_grenade_fire (ent, true);
				ent->client->grenade_blew_up = true;
			}

			if (ent->client->buttons & BUTTON_ATTACK)
				return;

			if (ent->client->grenade_blew_up)
			{
				if (level.time >= ent->client->grenade_time)
				{
					ent->client->ps.gunframe = 15;
					ent->client->grenade_blew_up = false;
				}
				else
				{
					return;
				}
			}
		}

		if (ent->client->ps.gunframe == 12)
		{
			ent->client->weapon_sound = 0;
			weapon_grenade_fire (ent, false);
		}

		if ((ent->client->ps.gunframe == 15) && (level.time < ent->client->grenade_time))
			return;

		ent->client->ps.gunframe++;

		if (ent->client->ps.gunframe == 16)
		{
			ent->client->grenade_time = 0;
			ent->client->weaponstate = WEAPON_READY;
		}
	}
}

/*
======================================================================

GRENADE LAUNCHER

======================================================================
*/

void weapon_grenadelauncher_fire (edict_t *ent)
{
	vec3_t	offset;
	vec3_t	forward, right;
	vec3_t	start;
	int		damage = 50;
	float	radius;

	radius = damage+40;
	if (is_quad)
		damage *= 4;

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 9;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	VectorSet(offset, 8, 8, ent->viewheight-8);
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	GrenLaunch_Source (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_grenade (ent, start, forward, damage, 600, 2.5, radius);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("gunner/gunatck3.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_GUNNER_GRENADE_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_GrenadeLauncher (edict_t *ent)
{
	static int	pause_frames[]	= {11, 20, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 3, 10, 21, 22, pause_frames, fire_frames, weapon_grenadelauncher_fire);
}

/*
======================================================================

ROCKET

======================================================================
*/

void Weapon_RocketLauncher_Fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;
	int		radius_damage;

	damage = 100 + (int)(random() * 20.0);
	radius_damage = 120;
	damage_radius = 120;
	if (is_quad)
	{
		damage *= 4;
		radius_damage *= 4;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	VectorSet(offset, 8, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_rocket (ent, start, forward, damage, 650, damage_radius, radius_damage);

	// send muzzle flash
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ_ROCKET | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_RocketLauncher (edict_t *ent)
{
	static int	pause_frames[]	= {25, 33, 42, 50, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 4, 12, 50, 54, pause_frames, fire_frames, Weapon_RocketLauncher_Fire);
}



void TankBlast_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	Left_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 800, effect, 3);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/tnkatck3.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_TankBlast_Fire (edict_t *ent)
{
	int		damage;

	damage = 30;

	TankBlast_Fire (ent, vec3_origin, damage, false, EF_BLASTER);
	ent->client->ps.gunframe++;
}

void Weapon_TankBlast (edict_t *ent)
{
	static int	pause_frames[]	= {7, 16, 0};
	static int	fire_frames[]	= {4, 0};

	Weapon_Generic (ent, 2, 6, 17, 18, pause_frames, fire_frames, Weapon_TankBlast_Fire);
}
void Weapon_TankRock_Fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;
	int		radius_damage;

	damage = 100 + (int)(random() * 20.0);
	radius_damage = 50+20;
	damage_radius = 50+20;
	if (is_quad)
	{
		damage *= 4;
		radius_damage *= 4;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 29;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	VectorSet(offset, 8, 8, ent->viewheight+4);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_rocket (ent, start, forward, 50, 550, damage_radius, radius_damage);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/tnkatck1.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_TankRock (edict_t *ent)
{
	static int	pause_frames[]	= {29, 38, 0};
	static int	fire_frames[]	= {14, 17, 20, 0};

	Weapon_Generic (ent, 2, 28, 39, 40, pause_frames, fire_frames, Weapon_TankRock_Fire);
}

void Weapon_IronRock_Fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;
	int		radius_damage;

	damage = 100 + (int)(random() * 20.0);
	radius_damage = 50+20;
	damage_radius = 50+20;
	if (is_quad)
	{
		damage *= 4;
		radius_damage *= 4;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 29;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	VectorSet(offset, 8, 8, ent->viewheight+4);
	GrenLaunch_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_rocket (ent, start, forward, 50, 550, damage_radius, radius_damage);

	gi.sound(ent, CHAN_AUTO, gi.soundindex ("chick/Chkatck2.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_IronRock (edict_t *ent)
{
	static int	pause_frames[]	= {34, 43, 0};
	static int	fire_frames[]	= {18, 0};

	Weapon_Generic (ent, 2, 33, 44, 45, pause_frames, fire_frames, Weapon_IronRock_Fire);
}

/*
======================================================================

BLASTER / HYPERBLASTER

======================================================================
*/

void HoverHyper_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("hover/hovatck1.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_HoverHyper_Fire (edict_t *ent)
{
	int		damage;

	damage = 1;

	HoverHyper_Fire (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
	ent->client->ps.gunframe++;
}

void Weapon_HoverHyper (edict_t *ent)
{
	static int	pause_frames[]	= {6, 15, 0};
	static int	fire_frames[]	= {3, 4, 0};

	Weapon_Generic (ent, 2, 5, 16, 17, pause_frames, fire_frames, Weapon_HoverHyper_Fire);
}

void FlyerHyper_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	Left_Flyer_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);
	ent->client->ps.gunframe++;

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flyer/flyatck3.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_FlyerHyper_Fire (edict_t *ent)
{
	int		damage;

	damage = 2;

	FlyerHyper_Fire (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
}

void Weapon_FlyerHyper (edict_t *ent)
{
	static int	pause_frames[]	= {6, 15, 0};
	static int	fire_frames[]	= {0};

	Weapon_Generic (ent, 1, 5, 16, 17, pause_frames, fire_frames, Weapon_FlyerHyper_Fire);
}

void FloaterHyper_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
	{
		if (ent->client->ps.gunframe == 7)
			ent->client->ps.gunframe = 6;
		else
			ent->client->ps.gunframe = 6;
	}
	else
		ent->client->ps.gunframe = 8;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("floater/Fltatck1.wav"), 1, ATTN_NORM, 0);

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_FloaterHyper_Fire (edict_t *ent)
{
	int		damage;

	damage = 2;

	FloaterHyper_Fire (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
}

void Weapon_FloaterHyper (edict_t *ent)
{
	static int	pause_frames[]	= {8, 17, 0};
	static int	fire_frames[]	= {6, 0};

	Weapon_Generic (ent, 2, 7, 18, 19, pause_frames, fire_frames, Weapon_FloaterHyper_Fire);
}



void Blaster_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 600, effect, 4);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("soldier/solatck2.wav"), 1, ATTN_NORM, 0);

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_Blaster_Fire (edict_t *ent)
{
	int		damage;

	damage = 5;

	Blaster_Fire (ent, vec3_origin, damage, false, EF_BLASTER);
	ent->client->ps.gunframe++;
}

void Weapon_Blaster (edict_t *ent)
{
	static int	pause_frames[]	= {11, 20, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 4, 10, 21, 24, pause_frames, fire_frames, Weapon_Blaster_Fire);
}

void MedicHyper_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 600, effect, 2);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("medic/medatck1.wav"), 1, ATTN_NORM, 0);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_MedicHyper_Fire (edict_t *ent)
{
	int		damage;

	damage = 5;

	MedicHyper_Fire (ent, vec3_origin, damage, false, EF_HYPERBLASTER);
	ent->client->ps.gunframe++;
}

void Weapon_MedicHyper (edict_t *ent)
{
	static int	pause_frames[]	= {18, 27, 0};
	static int	fire_frames[]	= {12, 13, 0};

	Weapon_Generic (ent, 2, 17, 28, 29, pause_frames, fire_frames, Weapon_MedicHyper_Fire);
}


void Weapon_HyperBlaster_Fire (edict_t *ent)
{
	float	rotation;
	vec3_t	offset;
	int		effect;
	int		damage;

	ent->client->weapon_sound = gi.soundindex("weapons/hyprbl1a.wav");

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
	}
	else
	{
		if (! ent->client->pers.inventory[ent->client->ammo_index] )
		{
			if (level.time >= ent->pain_debounce_time)
			{
				gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
				ent->pain_debounce_time = level.time + 1;
			}
		}
		else
		{
			rotation = (ent->client->ps.gunframe - 5) * 2*M_PI/6;
			offset[0] = -4 * sin(rotation);
			offset[1] = 0;
			offset[2] = 4 * cos(rotation);

			if ((ent->client->ps.gunframe == 6) || (ent->client->ps.gunframe == 9))
				effect = EF_HYPERBLASTER;
			else
				effect = 0;
			if (deathmatch->value)
				damage = 15;
			else
				damage = 20;
			Blaster_Fire (ent, offset, damage, true, effect);
			if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
				ent->client->pers.inventory[ent->client->ammo_index]--;

			ent->client->anim_priority = ANIM_ATTACK;
			if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
			{
				ent->s.frame = FRAME_crattak1 - 1;
				ent->client->anim_end = FRAME_crattak9;
			}
			else
			{
				ent->s.frame = FRAME_attack1 - 1;
				ent->client->anim_end = FRAME_attack8;
			}
		}

		ent->client->ps.gunframe++;
		if (ent->client->ps.gunframe == 12 && ent->client->pers.inventory[ent->client->ammo_index])
			ent->client->ps.gunframe = 6;
	}

	if (ent->client->ps.gunframe == 12)
	{
		gi.sound(ent, CHAN_AUTO, gi.soundindex("weapons/hyprbd1a.wav"), 1, ATTN_NORM, 0);
		ent->client->weapon_sound = 0;
	}

}

void Weapon_HyperBlaster (edict_t *ent)
{
	static int	pause_frames[]	= {0};
	static int	fire_frames[]	= {6, 7, 8, 9, 10, 11, 0};

	Weapon_Generic (ent, 5, 20, 49, 53, pause_frames, fire_frames, Weapon_HyperBlaster_Fire);
}

void GunChain_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 3;
	int			kick = 4;
	vec3_t		offset;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->ps.gunframe == 12)
		ent->client->ps.gunframe = 10;
	else
		ent->client->ps.gunframe = 10;

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 13;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bullet (ent, start, forward, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_GUNNERCHAIN);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("gunner/gunatck2.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_GUNNER_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_GunChain (edict_t *ent)
{
	static int	pause_frames[]	= {16, 25, 0};
	static int	fire_frames[]	= {10, 11, 12, 0};

	Weapon_Generic (ent, 2, 15, 26, 28, pause_frames, fire_frames, GunChain_Fire);
}

void InfMelee_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (5 + (rand() % 5));
	int			kick = 50;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee (ent, start, forward, 60, damage, kick, 1, MOD_INFMELEE);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_InfMelee (edict_t *ent)
{
	static int	pause_frames[]	= {16, 25, 0};
	static int	fire_frames[]	= {11, 0};

	Weapon_Generic (ent, 4, 15, 26, 28, pause_frames, fire_frames, InfMelee_Fire);
}

void SharkBite_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 45;
	int			kick = 0;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_brain (ent, start, forward, 100, damage, kick, 1, MOD_FLYSPIKES);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_SharkBite (edict_t *ent)
{
	static int	pause_frames[]	= {22, 31, 0};
	static int	fire_frames[]	= {15, 20, 0};

	Weapon_Generic (ent, 2, 21, 32, 33, pause_frames, fire_frames, SharkBite_Fire);
}


void SharkTail_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 35;
	int			kick = 0;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_brain (ent, start, forward, 100, damage, kick, 1, MOD_FLYSPIKES);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_SharkTail (edict_t *ent)
{
	static int	pause_frames[]	= {22, 31, 0};
	static int	fire_frames[]	= {12, 15, 0};

	Weapon_Generic (ent, 2, 21, 32, 33, pause_frames, fire_frames, SharkTail_Fire);
}


void FlyerMelee_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 5;
	int			kick = 0;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_flyer (ent, start, forward, 100, damage, kick, 1, MOD_FLYSPIKES);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("flyer/flyatck2.wav"), 1, ATTN_NORM, 0);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_FlyerMelee (edict_t *ent)
{
	static int	pause_frames[]	= {13, 22, 0};
	static int	fire_frames[]	= {5, 9, 0};

	Weapon_Generic (ent, 2, 12, 23, 24, pause_frames, fire_frames, FlyerMelee_Fire);
}

void GladMelee_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (5 + (rand() % 5));
	int			kick = 50;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_glad (ent, start, forward, 83, damage, kick, 1, MOD_GLADSPIKE);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_GladMelee (edict_t *ent)
{
	static int	pause_frames[]	= {10, 19, 0};
	static int	fire_frames[]	= {6, 0};

	Weapon_Generic (ent, 2, 9, 20, 21, pause_frames, fire_frames, GladMelee_Fire);
}

void BrainTentacles_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (5 + (rand() %5));
	int			kick = -600;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 70, damage, kick, 1, MOD_BRAINTENTACLE);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_BrainTentacles (edict_t *ent)
{
	static int	pause_frames[]	= {20, 29, 0};
	static int	fire_frames[]	= {9, 10, 11, 12, 0};

	Weapon_Generic (ent, 2, 19, 30, 31, pause_frames, fire_frames, BrainTentacles_Fire);
}

void BrainMelee_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (13 + (rand() %7));
	int			kick = 40;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_brain (ent, start, forward, 80, damage, kick, 1, MOD_BRAINCLAWS);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_BrainMelee (edict_t *ent)
{
	static int	pause_frames[]	= {22, 31, 0};
	static int	fire_frames[]	= {9, 14, 0};

	Weapon_Generic (ent, 2, 21, 32, 33, pause_frames, fire_frames, BrainMelee_Fire);
}

void MutantMelee_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (10 + (rand() %5));
	int			kick = 0;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_mutant2 (ent, start, forward, 80, damage, kick, 1, MOD_MUTSLASH);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("mutant/Mutatck1.wav"), 1, ATTN_NORM, 0);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_MutantMelee (edict_t *ent)
{
	static int	pause_frames[]	= {16, 25, 0};
	static int	fire_frames[]	= {9, 11, 0};

	Weapon_Generic (ent, 2, 15, 26, 27, pause_frames, fire_frames, MutantMelee_Fire);
}


void BerserkLeft_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (15 + (rand() % 6));
	int			kick = 400;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 57, damage, kick, 1, MOD_BERCLUB);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_BerserkLeft (edict_t *ent)
{
	static int	pause_frames[]	= {16, 25, 0};
	static int	fire_frames[]	= {9, 0};

	Weapon_Generic (ent, 2, 15, 26, 27, pause_frames, fire_frames, BerserkLeft_Fire);
}

void IronHand_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (10 + (rand() %6));
	int			kick = 100;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 100, damage, kick, 1, MOD_IRONHAND);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("chick/Chkatck3.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_IronHand (edict_t *ent)
{
	static int	pause_frames[]	= {13, 23, 0};
	static int	fire_frames[]	= {9, 0};

	Weapon_Generic (ent, 2, 12, 24, 25, pause_frames, fire_frames, IronHand_Fire);
}

void BerserkRight_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (5 + (rand() % 6));
	int			kick = 400;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 80, damage, kick, 1, MOD_BERSPIKE);
	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_BerserkRight (edict_t *ent)
{
	static int	pause_frames[]	= {10, 19, 0};
	static int	fire_frames[]	= {7, 0};

	Weapon_Generic (ent, 2, 9, 20, 21, pause_frames, fire_frames, BerserkRight_Fire);
}

void Boss1Chain_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 6;
	int			kick = 4;
	vec3_t		offset;
	vec3_t firevec; /*Added: Direction to fire*/

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 4;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	mod_GetLeadoffVec(ent, start, 5000, 1000, false, firevec);
	fire_bullet (ent, start, firevec, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_STANKCHAIN);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("infantry/infatck1.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_INFANTRY_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_Boss1Chain (edict_t *ent)
{
	static int	pause_frames[]	= {4, 13, 0};
	static int	fire_frames[]	= {2, 3, 0};

	Weapon_Generic (ent, 1, 3, 13, 14, pause_frames, fire_frames, Boss1Chain_Fire);
}





void Weapon_Boss2Rock_Fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;
	int		radius_damage;
	vec3_t  v;

	damage = 100 + (int)(random() * 20.0);
	radius_damage = 50+20;
	damage_radius = 50+20;
	if (is_quad)
	{
		damage *= 4;
		radius_damage *= 4;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 14;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	VectorSet(offset, 8, 8, ent->viewheight-12);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	v[PITCH] = ent->client->v_angle[PITCH];
	v[YAW]   = ent->client->v_angle[YAW] + 3;
	v[ROLL]  = ent->client->v_angle[ROLL];
	AngleVectors (v, forward, NULL, NULL);
	fire_rocket (ent, start, forward, 50, 500, damage_radius, radius_damage);
	v[PITCH] = ent->client->v_angle[PITCH];
	v[YAW]   = ent->client->v_angle[YAW] - 0;
	v[ROLL]  = ent->client->v_angle[ROLL];
	AngleVectors (v, forward, NULL, NULL);
	fire_rocket (ent, start, forward, 50, 500, damage_radius, radius_damage);
	v[PITCH] = ent->client->v_angle[PITCH];
	v[YAW]   = ent->client->v_angle[YAW] - 3;
	v[ROLL]  = ent->client->v_angle[ROLL];
	AngleVectors (v, forward, NULL, NULL);
	fire_rocket (ent, start, forward, 50, 500, damage_radius, radius_damage);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/rocket.wav"), 1, ATTN_NORM, 0);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/rocket.wav"), 1, ATTN_NORM, 0);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/rocket.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= 3;
}

void Weapon_Boss2Rock (edict_t *ent)
{
	static int	pause_frames[]	= {15, 24, 0};
	static int	fire_frames[]	= {8, 0};

	Weapon_Generic (ent, 2, 14, 25, 26, pause_frames, fire_frames, Weapon_Boss2Rock_Fire);
}





void Weapon_Boss1Rock_Fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;
	int		radius_damage;

	damage = 100 + (int)(random() * 20.0);
	radius_damage = 50+20;
	damage_radius = 50+20;
	if (is_quad)
	{
		damage *= 4;
		radius_damage *= 4;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 14;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	VectorSet(offset, 8, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_rocket (ent, start, forward, 50, 500, damage_radius, radius_damage);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("tank/rocket.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_Boss1Rock (edict_t *ent)
{
	static int	pause_frames[]	= {14, 23, 0};
	static int	fire_frames[]	= {3, 6, 9, 0};

	Weapon_Generic (ent, 2, 13, 24, 25, pause_frames, fire_frames, Weapon_Boss1Rock_Fire);
}


void InfChain_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 3;
	int			kick = 4;
	vec3_t		offset;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->ps.gunframe == 4)
		ent->client->ps.gunframe = 5;
	else
		ent->client->ps.gunframe = 5;

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 5;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bullet (ent, start, forward, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_INFCHAIN);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("infantry/infatck1.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_INFANTRY_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_InfChain (edict_t *ent)
{
	static int	pause_frames[]	= {6, 15, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 4, 5, 16, 18, pause_frames, fire_frames, InfChain_Fire);
}


void Boss2Chains_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 3;
	int			kick = 4;
	vec3_t		offset;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 5;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bullet (ent, start, forward, 6, 4, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_BOSS2CH1);
	fire_bullet (ent, start, forward, 6, 4, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_BOSS2CH2);
	ent->client->ps.gunframe++;

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("infantry/infatck1.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_INFANTRY_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= 2;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_Boss2Chains (edict_t *ent)
{
	static int	pause_frames[]	= {6, 15, 0};
	static int	fire_frames[]	= {2, 3, 4, 5, 0};

	Weapon_Generic (ent, 2, 5, 16, 17, pause_frames, fire_frames, Boss2Chains_Fire);
}

void Boss3Chains_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 3;
	int			kick = 4;
	vec3_t		offset;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 7;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bullet (ent, start, forward, 20, 10, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_BOSS2CH1);
	fire_bullet (ent, start, forward, 20, 10, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_BOSS2CH2);
	ent->client->ps.gunframe++;

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("berserk/xfire.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_INFANTRY_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= 2;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_Boss3Chains (edict_t *ent)
{
	static int	pause_frames[]	= {7, 16, 0};
	static int	fire_frames[]	= {2, 3, 4, 5, 6, 0};

	Weapon_Generic (ent, 2, 6, 17, 18, pause_frames, fire_frames, Boss3Chains_Fire);
}
/*
======================================================================

MACHINEGUN / CHAINGUN

======================================================================
*/

#define rndnum(y,z) ((random()*((z)-((y)+1)))+(y))

void TankMach_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 20;
	int			kick = 4;
	vec3_t		offset;
	float r;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->ps.gunframe == 2)
		ent->client->ps.gunframe = 3;
	else
		ent->client->ps.gunframe = 3;

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 5;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bullet (ent, start, forward, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_TANKMACHINEGUN);

	r = rndnum (1, 6);
	gi.sound (ent, CHAN_WEAPON, gi.soundindex("weapons/Tnkmc2.wav"), 1, ATTN_NORM, 0);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_TankMach (edict_t *ent)
{
	static int	pause_frames[]	= {4, 13, 0};
	static int	fire_frames[]	= {3, 0};

	Weapon_Generic (ent, 2, 3, 14, 15, pause_frames, fire_frames, TankMach_Fire);
}

void Machinegun_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = 2;
	int			kick = 4;
	vec3_t		offset;
	vec3_t		firevec;

	if (!(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe++;
		return;
	}

	if (ent->client->ps.gunframe == 4)
		ent->client->ps.gunframe = 5;
	else
		ent->client->ps.gunframe = 5;

	if (ent->client->pers.inventory[ent->client->ammo_index] < 1)
	{
		ent->client->ps.gunframe = 5;
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	mod_GetLeadoffVec(ent, start, 5000, 1000, false, firevec);
	fire_bullet (ent, start, firevec, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_SOLDMACHINEGUN);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("soldier/solatck3.wav"), 1, ATTN_NORM, 0);

	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_SOLDIER_MACHINEGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (int) (random()+0.25);
		ent->client->anim_end = FRAME_attack8;
	}
}

void Weapon_Machinegun (edict_t *ent)
{
	static int	pause_frames[]	= {6, 15, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 4, 5, 16, 19, pause_frames, fire_frames, Machinegun_Fire);
}

void Chaingun_Fire (edict_t *ent)
{
	int			i;
	int			shots;
	vec3_t		start;
	vec3_t		forward, right, up;
	float		r, u;
	vec3_t		offset;
	int			damage;
	int			kick = 2;

	if (deathmatch->value)
		damage = 6;
	else
		damage = 8;

	if (ent->client->ps.gunframe == 5)
		gi.sound(ent, CHAN_AUTO, gi.soundindex("weapons/chngnu1a.wav"), 1, ATTN_IDLE, 0);

	if ((ent->client->ps.gunframe == 14) && !(ent->client->buttons & BUTTON_ATTACK))
	{
		ent->client->ps.gunframe = 32;
		ent->client->weapon_sound = 0;
		return;
	}
	else if ((ent->client->ps.gunframe == 21) && (ent->client->buttons & BUTTON_ATTACK)
		&& ent->client->pers.inventory[ent->client->ammo_index])
	{
		ent->client->ps.gunframe = 15;
	}
	else
	{
		ent->client->ps.gunframe++;
	}

	if (ent->client->ps.gunframe == 22)
	{
		ent->client->weapon_sound = 0;
		gi.sound(ent, CHAN_AUTO, gi.soundindex("weapons/chngnd1a.wav"), 1, ATTN_IDLE, 0);
	}
	else
	{
		ent->client->weapon_sound = gi.soundindex("weapons/chngnl1a.wav");
	}

	ent->client->anim_priority = ANIM_ATTACK;
	if (ent->client->ps.pmove.pm_flags & PMF_DUCKED)
	{
		ent->s.frame = FRAME_crattak1 - (ent->client->ps.gunframe & 1);
		ent->client->anim_end = FRAME_crattak9;
	}
	else
	{
		ent->s.frame = FRAME_attack1 - (ent->client->ps.gunframe & 1);
		ent->client->anim_end = FRAME_attack8;
	}

	if (ent->client->ps.gunframe <= 9)
		shots = 1;
	else if (ent->client->ps.gunframe <= 14)
	{
		if (ent->client->buttons & BUTTON_ATTACK)
			shots = 2;
		else
			shots = 1;
	}
	else
		shots = 3;

	if (ent->client->pers.inventory[ent->client->ammo_index] < shots)
		shots = ent->client->pers.inventory[ent->client->ammo_index];

	if (!shots)
	{
		if (level.time >= ent->pain_debounce_time)
		{
			gi.sound(ent, CHAN_VOICE, gi.soundindex("weapons/noammo.wav"), 1, ATTN_NORM, 0);
			ent->pain_debounce_time = level.time + 1;
		}
		return;
	}

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	for (i=0 ; i<3 ; i++)
	{
		ent->client->kick_origin[i] = crandom() * 0.35;
		ent->client->kick_angles[i] = crandom() * 0.7;
	}

	for (i=0 ; i<shots ; i++)
	{
		// get start / end positions
		AngleVectors (ent->client->v_angle, forward, right, up);
		r = 7 + crandom()*4;
		u = crandom()*4;
		VectorSet(offset, 0, r, u + ent->viewheight-8);
		P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

		fire_bullet (ent, start, forward, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, MOD_CHAINGUN);
	}

	// send muzzle flash
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte ((MZ_CHAINGUN1 + shots - 1) | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= shots;
}


void Weapon_Chaingun (edict_t *ent)
{
	static int	pause_frames[]	= {38, 43, 51, 61, 0};
	static int	fire_frames[]	= {5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 0};

	Weapon_Generic (ent, 4, 31, 61, 64, pause_frames, fire_frames, Chaingun_Fire);
}


/*
======================================================================

SHOTGUN / SUPERSHOTGUN

======================================================================
*/

void weapon_shotgun_fire (edict_t *ent)
{
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		offset;
	int			damage = 2;
	int			kick = 1;

	if (ent->client->ps.gunframe == 9)
	{
		ent->client->ps.gunframe++;
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -2;

	VectorSet(offset, 0, 8,  ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	fire_shotgun (ent, start, forward, damage, kick, 500, 500, DEFAULT_DEATHMATCH_SHOTGUN_COUNT, MOD_SOLDSHOTGUN);
	
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("soldier/solatck1.wav"), 1, ATTN_NORM, 0);

	// send muzzle flash
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ2_SOLDIER_SHOTGUN_1 | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	ent->client->ps.gunframe++;
	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}

void Weapon_Shotgun (edict_t *ent)
{
	static int	pause_frames[]	= {15, 24, 0};
	static int	fire_frames[]	= {5, 9, 0};

	Weapon_Generic (ent, 4, 14, 25, 28, pause_frames, fire_frames, weapon_shotgun_fire);
}


void weapon_supershotgun_fire (edict_t *ent)
{
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		offset;
	vec3_t		v;
	int			damage = 6;
	int			kick = 12;

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -2;

	VectorSet(offset, 0, 8,  ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	v[PITCH] = ent->client->v_angle[PITCH];
	v[YAW]   = ent->client->v_angle[YAW] - 5;
	v[ROLL]  = ent->client->v_angle[ROLL];
	AngleVectors (v, forward, NULL, NULL);
	fire_shotgun (ent, start, forward, damage, kick, DEFAULT_SHOTGUN_HSPREAD, DEFAULT_SHOTGUN_VSPREAD, DEFAULT_SSHOTGUN_COUNT/2, MOD_SSHOTGUN);
	v[YAW]   = ent->client->v_angle[YAW] + 5;
	AngleVectors (v, forward, NULL, NULL);
	fire_shotgun (ent, start, forward, damage, kick, DEFAULT_SHOTGUN_HSPREAD, DEFAULT_SHOTGUN_VSPREAD, DEFAULT_SSHOTGUN_COUNT/2, MOD_SSHOTGUN);

	// send muzzle flash
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (ent-g_edicts);
	gi.WriteByte (MZ_SSHOTGUN | is_silenced);
	gi.multicast (ent->s.origin, MULTICAST_PVS);

	ent->client->ps.gunframe++;
	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= 2;
}

void Weapon_SuperShotgun (edict_t *ent)
{
	static int	pause_frames[]	= {29, 42, 57, 0};
	static int	fire_frames[]	= {7, 0};

	Weapon_Generic (ent, 6, 17, 57, 61, pause_frames, fire_frames, weapon_supershotgun_fire);
}



/*
======================================================================

RAILGUN

======================================================================
*/

void weapon_railgun_fire (edict_t *ent)
{
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		offset;
	int			damage;
	int			kick;

	damage = 50;
	kick = 100;

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -3, ent->client->kick_origin);
	ent->client->kick_angles[0] = -3;

	VectorSet(offset, 0, 7,  ent->viewheight-8);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_rail (ent, start, forward, damage, kick, MOD_GLADRAIL);

	ent->client->ps.gunframe++;
	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}


void Weapon_Railgun (edict_t *ent)
{
	static int	pause_frames[]	= {16, 25, 0};
	static int	fire_frames[]	= {7, 0};

	Weapon_Generic (ent, 2, 15, 26, 27, pause_frames, fire_frames, weapon_railgun_fire);
}

void Weapon_MakronRail_Fire (edict_t *ent)
{
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		offset;
	int			damage;
	int			kick;

	damage = 50;
	kick = 100;

	if (is_quad)
	{
		damage *= 4;
		kick *= 4;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -3, ent->client->kick_origin);
	ent->client->kick_angles[0] = -3;

	VectorSet(offset, 0, 7,  ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);
	fire_rail (ent, start, forward, damage, kick, MOD_GLADRAIL);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("weapons/RAILGF1A.WAV"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;
	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index]--;
}


void Weapon_MakronRail (edict_t *ent)
{
	static int	pause_frames[]	= {19, 28, 0};
	static int	fire_frames[]	= {11, 0};

	Weapon_Generic (ent, 2, 18, 29, 30, pause_frames, fire_frames, Weapon_MakronRail_Fire);
}


/*
======================================================================

BFG10K

======================================================================
*/

void weapon_bfg_fire (edict_t *ent)
{
	vec3_t	offset, start;
	vec3_t	forward, right;
	int		damage;
	float	damage_radius;

	// cells can go down during windup (from power armor hits), so
	// check again and abort firing if we don't have enough now
	if (ent->client->pers.inventory[ent->client->ammo_index] < 20)
	{
		return;
	}

	AngleVectors (ent->client->v_angle, forward, right, NULL);

	VectorScale (forward, -2, ent->client->kick_origin);

	// make a big pitch kick with an inverse fall
	ent->client->v_dmg_pitch = -5;
	ent->client->v_dmg_roll = crandom()*2;
	ent->client->v_dmg_time = level.time + DAMAGE_TIME;

	VectorSet(offset, 8, 8, ent->viewheight-8);
	Left_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);
	fire_bfg (ent, start, forward, 50, 300, 300);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("makron/bfg_fire.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] -= 20;
}

void Weapon_BFG (edict_t *ent) 
{
	static int	pause_frames[]	= {10, 19, 0};
	static int	fire_frames[]	= {5, 0};

	Weapon_Generic (ent, 2, 9, 20, 21, pause_frames, fire_frames, weapon_bfg_fire);
} 


void MakronBlast_Fire (edict_t *ent, vec3_t g_offset, int damage, qboolean hyper, int effect)
{
	vec3_t	forward, right;
	vec3_t	start;
	vec3_t	offset;

	if ( ((ent->client->latched_buttons|ent->client->buttons) & BUTTON_ATTACK) )
	{
		if (ent->client->ps.gunframe == 2)
			ent->client->ps.gunframe = 3;
		else
			ent->client->ps.gunframe = 2;
	}

	if (is_quad)
		damage *= 4;
	AngleVectors (ent->client->v_angle, forward, right, NULL);
	VectorSet(offset, 24, 8, ent->viewheight-8);
	VectorAdd (offset, g_offset, offset);
	P_ProjectSource (ent->client, ent->s.origin, offset, forward, right, start);

	VectorScale (forward, -2, ent->client->kick_origin);
	ent->client->kick_angles[0] = -1;

	fire_blaster (ent, start, forward, damage, 1000, effect, 2);

	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("makron/blaster.wav"), 1, ATTN_NORM, 0);

	if (! ( (int)dmflags->value & DF_INFINITE_AMMO ) )
		ent->client->pers.inventory[ent->client->ammo_index] --;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_MakronBlast_Fire (edict_t *ent)
{
	int		damage;

	damage = 15;

	MakronBlast_Fire (ent, vec3_origin, damage, false, EF_BLASTER);
	ent->client->ps.gunframe++;
}

void Weapon_MakronBlast (edict_t *ent)
{
	static int	pause_frames[]	= {4, 13, 0};
	static int	fire_frames[]	= {3, 0};

	Weapon_Generic (ent, 2, 3, 14, 15, pause_frames, fire_frames, Weapon_MakronBlast_Fire);
}

//======================================================================

void FloaterMelee1_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (30 + rand() % 6);
	int			kick = -50;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 65, damage, kick, 1, MOD_BRAINTENTACLE);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("floater/Fltatck3.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_FloaterMelee1 (edict_t *ent)
{
	static int	pause_frames[]	= {13, 22, 0};
	static int	fire_frames[]	= {10, 0};

	Weapon_Generic (ent, 2, 12, 23, 24, pause_frames, fire_frames, FloaterMelee1_Fire);
}

void FloaterMelee2_Fire (edict_t *ent)
{
	int	i;
	vec3_t		start;
	vec3_t		forward, right;
	vec3_t		angles;
	int			damage = (35 + rand() % 6);
	int			kick = -50;
	vec3_t		offset;

	// get start / end positions
	VectorAdd (ent->client->v_angle, ent->client->kick_angles, angles);
	AngleVectors (angles, forward, right, NULL);
	VectorSet(offset, 0, 8, ent->viewheight-8);
	Middle_Source (ent->client, ent->s.origin, offset, forward, right, start);

	fire_player_melee_berserk (ent, start, forward, 80, damage, kick, 1, MOD_BRAINTENTACLE);
	gi.sound(ent, CHAN_WEAPON, gi.soundindex ("floater/Fltatck2.wav"), 1, ATTN_NORM, 0);

	ent->client->ps.gunframe++;

	PlayerNoise(ent, start, PNOISE_WEAPON);
}

void Weapon_FloaterMelee2 (edict_t *ent)
{
	static int	pause_frames[]	= {12, 21, 0};
	static int	fire_frames[]	= {10, 0};

	Weapon_Generic (ent, 2, 11, 22, 23, pause_frames, fire_frames, FloaterMelee2_Fire);
}